# --- START LICENSE (optional)
# --- END LICENSE
# --- START IMPORT SECTION
import logging

from hamilton import contrib

logger = logging.getLogger(__name__)

with contrib.catch_import_errors(__name__, __file__, logger):
    # non-hamilton imports go here
    pass

# hamilton imports go here; check for required version if need be.

# --- END IMPORT SECTION

# --- START HAMILTON DATAFLOW


# --- END HAMILTON DATAFLOW
# --- START MAIN CODE
if __name__ == "__main__":
    # Code to create an imaging showing on DAG workflow.
    # run as a script to test Hamilton's execution
    import __init__ as MODULE_NAME

    from hamilton import base, driver

    dr = driver.Driver(
        {},  # CONFIG: fill as appropriate
        MODULE_NAME,
        adapter=base.DefaultAdapter(),
    )
    # saves to current working directory creating dag.png.
    dr.display_all_functions("dag", {"format": "png", "view": False})
# --- END MAIN CODE



from typing import Any, Type

from hamilton import registry

try:
    import ibis
    import ibis.expr.types as ir
except ImportError:
    raise NotImplementedError("Ibis is not installed.")

from hamilton.data_quality import base, default_validators

DATAFRAME_TYPE = ir.Table
COLUMN_TYPE = ir.Column


@registry.get_column.register(ir.Table)
def get_column_ibis(df: ir.Table, column_name: str) -> ir.Column:
    return df[column_name]


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("ibis", DATAFRAME_TYPE, COLUMN_TYPE)


@registry.fill_with_scalar.register(ir.Table)
def fill_with_scalar_ibis(df: ir.Table, column_name: str, scalar_value: Any) -> ir.Table:
    raise NotImplementedError(
        "As of Hamilton version 1.5.1, filling Ibis columns with scalars isn't supported. This feature is actively being developed"
    )


register_types()


class SchemaValidatorIbis(base.DataValidator):
    def __init__(self, schema: ibis.expr.schema.Schema, importance: str):
        """`schema` is an ordered mapping"""
        super(SchemaValidatorIbis, self).__init__(importance)
        self.schema = schema

    def name(self) -> str:
        return f"{self.arg()}_validator"

    @classmethod
    def arg(cls) -> str:
        return "schema"

    @classmethod
    def applies_to(cls, datatype: Type[Type]) -> bool:
        return issubclass(datatype, ir.Table)

    def description(self) -> str:
        return "Validates that ibis Table expression matches schema."

    def validate(self, data: ir.Table) -> base.ValidationResult:
        passes = data.schema().equals(self.schema)
        if passes:
            message = "Data passes Ibis schema check"
            diagnostics = {"schema": self.schema.fields}
        else:
            message = "Data failed Ibis schema check"
            diagnostics = {"schema": self.schema.fields}
        return base.ValidationResult(
            passes=passes,
            message=message,
            diagnostics=diagnostics,
        )


def register_validators():
    default_validators.AVAILABLE_DEFAULT_VALIDATORS.append(SchemaValidatorIbis)


register_validators()



import dataclasses
from io import BytesIO, IOBase, TextIOWrapper
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Collection,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Type,
    Union,
)

try:
    from xlsxwriter.workbook import Workbook
except ImportError:
    Workbook = Type

try:
    import polars as pl
except ImportError:
    raise NotImplementedError("Polars is not installed.")


# for polars <0.16.0 we need to determine whether type_aliases exist.
has_alias = False
if hasattr(pl, "type_aliases"):
    has_alias = True

# for polars 0.18.0 we need to check what to do.
if has_alias and hasattr(pl.type_aliases, "CsvEncoding"):
    from polars.type_aliases import CsvEncoding, SchemaDefinition
else:
    CsvEncoding = Type
if has_alias and hasattr(pl.type_aliases, "CsvQuoteStyle"):
    from polars.type_aliases import CsvQuoteStyle
else:
    CsvQuoteStyle = Type
if has_alias and hasattr(pl.type_aliases, "IpcCompression"):
    from polars.type_aliases import IpcCompression
else:
    IpcCompression = Type

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader, DataSaver

DATAFRAME_TYPE = pl.DataFrame
COLUMN_TYPE = pl.Series


@registry.get_column.register(pl.DataFrame)
def get_column_polars(df: pl.DataFrame, column_name: str) -> pl.Series:
    return df[column_name]


@registry.fill_with_scalar.register(pl.DataFrame)
def fill_with_scalar_polars(df: pl.DataFrame, column_name: str, scalar_value: Any) -> pl.DataFrame:
    if not isinstance(scalar_value, pl.Series):
        scalar_value = [scalar_value]
    return df.with_column(pl.Series(name=column_name, values=scalar_value))


@dataclasses.dataclass
class PolarsCSVReader(DataLoader):
    """Class specifically to handle loading CSV files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_csv.html
    """

    file: Union[str, TextIO, BytesIO, Path, BinaryIO, bytes]
    # kwargs:
    has_header: bool = True
    include_header: bool = True
    columns: Union[Sequence[int], Sequence[str]] = None
    new_columns: Sequence[str] = None
    separator: str = ","
    comment_char: str = None
    quote_char: str = '"'
    skip_rows: int = 0
    dtypes: Union[Mapping[str, Any], Sequence[Any]] = None
    null_values: Union[str, Sequence[str], Dict[str, str]] = None
    missing_utf8_is_empty_string: bool = False
    ignore_errors: bool = False
    try_parse_dates: bool = False
    n_threads: int = None
    infer_schema_length: int = 100
    batch_size: int = 8192
    n_rows: int = None
    encoding: Union[CsvEncoding, str] = "utf8"
    low_memory: bool = False
    rechunk: bool = True
    use_pyarrow: bool = False
    storage_options: Dict[str, Any] = None
    skip_rows_after_header: int = 0
    row_count_name: str = None
    row_count_offset: int = 0
    sample_size: int = 1024
    eol_char: str = "\n"
    raise_if_empty: bool = True

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.has_header is not None:
            kwargs["has_header"] = self.has_header
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.new_columns is not None:
            kwargs["new_columns"] = self.new_columns
        if self.separator is not None:
            kwargs["separator"] = self.separator
        if self.comment_char is not None:
            kwargs["comment_char"] = self.comment_char
        if self.quote_char is not None:
            kwargs["quote_char"] = self.quote_char
        if self.skip_rows is not None:
            kwargs["skip_rows"] = self.skip_rows
        if self.dtypes is not None:
            kwargs["dtypes"] = self.dtypes
        if self.null_values is not None:
            kwargs["null_values"] = self.null_values
        if self.missing_utf8_is_empty_string is not None:
            kwargs["missing_utf8_is_empty_string"] = self.missing_utf8_is_empty_string
        if self.ignore_errors is not None:
            kwargs["ignore_errors"] = self.ignore_errors
        if self.try_parse_dates is not None:
            kwargs["try_parse_dates"] = self.try_parse_dates
        if self.n_threads is not None:
            kwargs["n_threads"] = self.n_threads
        if self.infer_schema_length is not None:
            kwargs["infer_schema_length"] = self.infer_schema_length
        if self.batch_size is not None:
            kwargs["batch_size"] = self.batch_size
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        if self.encoding is not None:
            kwargs["encoding"] = self.encoding
        if self.low_memory is not None:
            kwargs["low_memory"] = self.low_memory
        if self.rechunk is not None:
            kwargs["rechunk"] = self.rechunk
        if self.use_pyarrow is not None:
            kwargs["use_pyarrow"] = self.use_pyarrow
        if self.storage_options is not None:
            kwargs["storage_options"] = self.storage_options
        if self.skip_rows_after_header is not None:
            kwargs["skip_rows_after_header"] = self.skip_rows_after_header
        if self.row_count_name is not None:
            kwargs["row_count_name"] = self.row_count_name
        if self.row_count_offset is not None:
            kwargs["row_count_offset"] = self.row_count_offset
        if self.eol_char is not None:
            kwargs["eol_char"] = self.eol_char
        if self.raise_if_empty is not None:
            kwargs["raise_if_empty"] = self.raise_if_empty
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_csv(self.file, **self._get_loading_kwargs())

        metadata = utils.get_file_and_dataframe_metadata(self.file, df)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "csv"


@dataclasses.dataclass
class PolarsCSVWriter(DataSaver):
    """Class specifically to handle saving CSV files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_csv.html
    """

    file: Union[BytesIO, TextIOWrapper, str, Path]
    # kwargs:
    has_header: bool = True
    separator: str = ","
    line_terminator: str = "\n"
    quote: str = '"'
    batch_size: int = 1024
    datetime_format: str = None
    date_format: str = None
    time_format: str = None
    float_precision: int = None
    null_value: str = None
    quote_style: CsvQuoteStyle = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.separator is not None:
            kwargs["separator"] = self.separator
        if self.has_header is not None:
            kwargs["has_header"] = self.has_header
        if self.separator is not None:
            kwargs["separator"] = self.separator
        if self.line_terminator is not None:
            kwargs["line_terminator"] = self.line_terminator
        if self.quote is not None:
            kwargs["quote"] = self.quote
        if self.batch_size is not None:
            kwargs["batch_size"] = self.batch_size
        if self.datetime_format is not None:
            kwargs["datetime_format"] = self.datetime_format
        if self.date_format is not None:
            kwargs["date_format"] = self.date_format
        if self.time_format is not None:
            kwargs["time_format"] = self.time_format
        if self.float_precision is not None:
            kwargs["float_precision"] = self.float_precision
        if self.null_value is not None:
            kwargs["null_value"] = self.null_value
        if self.quote_style is not None:
            kwargs["quote_style"] = self.quote_style
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()
        data.write_csv(self.file, **self._get_saving_kwargs())
        return utils.get_file_and_dataframe_metadata(self.file, data)

    @classmethod
    def name(cls) -> str:
        return "csv"


@dataclasses.dataclass
class PolarsParquetReader(DataLoader):
    """Class specifically to handle loading parquet files with polars
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_parquet.html
    """

    file: Union[str, TextIO, BytesIO, Path, BinaryIO, bytes]
    # kwargs:
    columns: Union[List[int], List[str]] = None
    n_rows: int = None
    use_pyarrow: bool = False
    memory_map: bool = True
    storage_options: Dict[str, Any] = None
    parallel: Any = "auto"
    row_count_name: str = None
    row_count_offset: int = 0
    low_memory: bool = False
    pyarrow_options: Dict[str, Any] = None
    use_statistics: bool = True
    rechunk: bool = True

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        if self.use_pyarrow is not None:
            kwargs["use_pyarrow"] = self.use_pyarrow
        if self.memory_map is not None:
            kwargs["memory_map"] = self.memory_map
        if self.storage_options is not None:
            kwargs["storage_options"] = self.storage_options
        if self.parallel is not None:
            kwargs["parallel"] = self.parallel
        if self.row_count_name is not None:
            kwargs["row_count_name"] = self.row_count_name
        if self.row_count_offset is not None:
            kwargs["row_count_offset"] = self.row_count_offset
        if self.low_memory is not None:
            kwargs["low_memory"] = self.low_memory
        if self.pyarrow_options is not None:
            kwargs["pyarrow_options"] = self.pyarrow_options
        if self.use_statistics is not None:
            kwargs["use_statistics"] = self.use_statistics
        if self.rechunk is not None:
            kwargs["rechunk"] = self.rechunk
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_parquet(self.file, **self._get_loading_kwargs())
        metadata = utils.get_file_and_dataframe_metadata(self.file, df)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "parquet"


@dataclasses.dataclass
class PolarsParquetWriter(DataSaver):
    """Class specifically to handle saving CSV files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_parquet.html
    """

    file: Union[BytesIO, TextIOWrapper, str, Path]
    # kwargs:
    compression: Any = "zstd"
    compression_level: int = None
    statistics: bool = False
    row_group_size: int = None
    use_pyarrow: bool = False
    pyarrow_options: Dict[str, Any] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.compression is not None:
            kwargs["compression"] = self.compression
        if self.compression is not None:
            kwargs["compression_level"] = self.compression_level
        if self.compression is not None:
            kwargs["statistics"] = self.statistics
        if self.compression is not None:
            kwargs["row_group_size"] = self.row_group_size
        if self.compression is not None:
            kwargs["use_pyarrow"] = self.use_pyarrow
        if self.compression is not None:
            kwargs["pyarrow_options"] = self.pyarrow_options
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()

        data.write_parquet(self.file, **self._get_saving_kwargs())

        return utils.get_file_and_dataframe_metadata(self.file, data)

    @classmethod
    def name(cls) -> str:
        return "parquet"


@dataclasses.dataclass
class PolarsFeatherReader(DataLoader):
    """
    Class specifically to handle loading Feather/Arrow IPC files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_ipc.html
    """

    source: Union[str, BinaryIO, BytesIO, Path, bytes]
    # kwargs:
    columns: Optional[Union[List[str], List[int]]] = None
    n_rows: Optional[int] = None
    use_pyarrow: bool = False
    memory_map: bool = True
    storage_options: Optional[Dict[str, Any]] = None
    row_count_name: Optional[str] = None
    row_count_offset: int = 0
    rechunk: bool = True

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        if self.use_pyarrow is not None:
            kwargs["use_pyarrow"] = self.use_pyarrow
        if self.memory_map is not None:
            kwargs["memory_map"] = self.memory_map
        if self.storage_options is not None:
            kwargs["storage_options"] = self.storage_options
        if self.row_count_name is not None:
            kwargs["row_count_name"] = self.row_count_name
        if self.row_count_offset is not None:
            kwargs["row_count_offset"] = self.row_count_offset
        if self.rechunk is not None:
            kwargs["rechunk"] = self.rechunk
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_ipc(self.source, **self._get_loading_kwargs())
        metadata = utils.get_file_metadata(self.source)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "feather"


@dataclasses.dataclass
class PolarsFeatherWriter(DataSaver):
    """
    Class specifically to handle saving Feather/Arrow IPC files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_ipc.html
    """

    file: Optional[Union[BinaryIO, BytesIO, str, Path]] = None
    # kwargs:
    compression: IpcCompression = "uncompressed"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.compression is not None:
            kwargs["compression"] = self.compression
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()
        data.write_ipc(self.file, **self._get_saving_kwargs())
        return utils.get_file_and_dataframe_metadata(self.file, data)

    @classmethod
    def name(cls) -> str:
        return "feather"


@dataclasses.dataclass
class PolarsAvroReader(DataLoader):
    """Class specifically to handle loading Avro files with polars
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_avro.html
    """

    file: Union[str, TextIO, BytesIO, Path, BinaryIO, bytes]
    # kwargs:
    columns: Union[List[int], List[str], None] = None
    n_rows: Union[int, None] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_avro(self.file, **self._get_loading_kwargs())
        metadata = utils.get_file_and_dataframe_metadata(self.file, df)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "avro"


@dataclasses.dataclass
class PolarsAvroWriter(DataSaver):
    """Class specifically to handle saving Avro files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_avro.html
    """

    file: Union[BytesIO, TextIOWrapper, str, Path]
    # kwargs:
    compression: Any = "uncompressed"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.compression is not None:
            kwargs["compression"] = self.compression
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()

        data.write_avro(self.file, **self._get_saving_kwargs())
        return utils.get_file_and_dataframe_metadata(self.file, data)

    @classmethod
    def name(cls) -> str:
        return "avro"


@dataclasses.dataclass
class PolarsJSONReader(DataLoader):
    """
    Class specifically to handle loading JSON files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_json.html
    """

    source: Union[str, Path, IOBase, bytes]
    schema: SchemaDefinition = None
    schema_overrides: SchemaDefinition = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.schema is not None:
            kwargs["schema"] = self.schema
        if self.schema_overrides is not None:
            kwargs["schema_overrides"] = self.schema_overrides
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_json(self.source, **self._get_loading_kwargs())
        metadata = utils.get_file_metadata(self.source)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "json"


@dataclasses.dataclass
class PolarsJSONWriter(DataSaver):
    """
    Class specifically to handle saving JSON files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_json.html
    """

    file: Union[IOBase, str, Path]
    pretty: bool = False
    row_oriented: bool = False

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.pretty is not None:
            kwargs["pretty"] = self.pretty
        if self.row_oriented is not None:
            kwargs["row_oriented"] = self.row_oriented
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()

        data.write_json(self.file, **self._get_saving_kwargs())
        return utils.get_file_and_dataframe_metadata(self.file, data)

    @classmethod
    def name(cls) -> str:
        return "json"


@dataclasses.dataclass
class PolarsSpreadsheetReader(DataLoader):
    """
    Class specifically to handle loading Spreadsheet files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_excel.html
    """

    source: Union[str, Path, IOBase, bytes]
    # kwargs:
    sheet_id: Union[int, Sequence[int], None] = None
    sheet_name: Union[str, List[str], Tuple[str], None] = None
    engine: Literal["xlsx2csv", "openpyxl", "pyxlsb", "odf", "xlrd", "xlsxwriter"] = "xlsx2csv"
    engine_options: Union[Dict[str, Any], None] = None
    read_options: Union[Dict[str, Any], None] = None
    schema_overrides: Union[Dict[str, Any], None] = None
    raise_if_empty: bool = True

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.sheet_id is not None:
            kwargs["sheet_id"] = self.sheet_id
        if self.sheet_name is not None:
            kwargs["sheet_name"] = self.sheet_name
        if self.engine is not None:
            kwargs["engine"] = self.engine
        if self.engine_options is not None:
            kwargs["engine_options"] = self.engine_options
        if self.read_options is not None:
            kwargs["read_options"] = self.read_options
        if self.schema_overrides is not None:
            kwargs["schema_overrides"] = self.schema_overrides
        if self.raise_if_empty is not None:
            kwargs["raise_if_empty"] = self.raise_if_empty
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_excel(self.source, **self._get_loading_kwargs())
        metadata = utils.get_file_metadata(self.source)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "spreadsheet"


@dataclasses.dataclass
class PolarsSpreadsheetWriter(DataSaver):
    """
    Class specifically to handle saving Spreadsheet files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.DataFrame.write_excel.html
    """

    # importing here because this is where it's used. Can move later.
    # but yeah the polars type aliases weren't resolving well in python 3.9
    # so stripped/reduced them appropriately.
    from polars.datatypes import DataType, DataTypeClass
    from polars.type_aliases import ColumnTotalsDefinition, RowTotalsDefinition

    workbook: Union[Workbook, BytesIO, Path, str]
    worksheet: Union[str, None] = None
    # kwargs:
    position: Union[Tuple[int, int], str] = "A1"
    table_style: Union[str, Dict[str, Any], None] = None
    table_name: Union[str, None] = None
    column_formats: Union[
        Mapping[Union[str, Tuple[str, ...]], Union[str, Mapping[str, str]]], None
    ] = None
    dtype_formats: Union[Dict[Union[DataType, DataTypeClass], str], None] = None
    conditional_formats: Union[
        Mapping[
            Union[str, Collection[str]],
            Union[str, Union[Mapping[str, Any], Sequence[Union[str, Mapping[str, Any]]]]],
        ],
        None,
    ] = None
    header_format: Union[Dict[str, Any], None] = None
    column_totals: Union[ColumnTotalsDefinition, None] = None
    column_widths: Union[Mapping[str, Union[Tuple[str, ...], int]], int, None] = None
    row_totals: Union[RowTotalsDefinition, None] = None
    row_heights: Union[Dict[Union[int, Tuple[int, ...]], int], int, None] = None
    sparklines: Union[Dict[str, Union[Sequence[str], Dict[str, Any]]], None] = None
    formulas: Union[Dict[str, Union[str, Dict[str, str]]], None] = None
    float_precision: int = 3
    include_header: bool = True
    autofilter: bool = True
    autofit: bool = False
    hidden_columns: Union[Sequence[str], str, None] = None
    hide_gridlines: bool = None
    sheet_zoom: Union[int, None] = None
    freeze_panes: Union[
        str, Tuple[int, int], Tuple[str, int, int], Tuple[int, int, int, int], None
    ] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.position is not None:
            kwargs["position"] = self.position
        if self.table_style is not None:
            kwargs["table_style"] = self.table_style
        if self.table_name is not None:
            kwargs["table_name"] = self.table_name
        if self.column_formats is not None:
            kwargs["column_formats"] = self.column_formats
        if self.dtype_formats is not None:
            kwargs["dtype_formats"] = self.dtype_formats
        if self.conditional_formats is not None:
            kwargs["conditional_formats"] = self.conditional_formats
        if self.header_format is not None:
            kwargs["header_format"] = self.header_format
        if self.column_totals is not None:
            kwargs["column_totals"] = self.column_totals
        if self.column_widths is not None:
            kwargs["column_widths"] = self.column_widths
        if self.row_totals is not None:
            kwargs["row_totals"] = self.row_totals
        if self.row_heights is not None:
            kwargs["row_heights"] = self.row_heights
        if self.sparklines is not None:
            kwargs["sparklines"] = self.sparklines
        if self.formulas is not None:
            kwargs["formulas"] = self.formulas
        if self.float_precision is not None:
            kwargs["float_precision"] = self.float_precision
        if self.include_header is not None:
            kwargs["include_header"] = self.include_header
        if self.autofilter is not None:
            kwargs["autofilter"] = self.autofilter
        if self.autofit is not None:
            kwargs["autofit"] = self.autofit
        if self.hidden_columns is not None:
            kwargs["hidden_columns"] = self.hidden_columns
        if self.hide_gridlines is not None:
            kwargs["hide_gridlines"] = self.hide_gridlines
        if self.sheet_zoom is not None:
            kwargs["sheet_zoom"] = self.sheet_zoom
        if self.freeze_panes is not None:
            kwargs["freeze_panes"] = self.freeze_panes
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()

        data.write_excel(self.workbook, self.worksheet, **self._get_saving_kwargs())
        return utils.get_file_and_dataframe_metadata(self.workbook, data)

    @classmethod
    def name(cls) -> str:
        return "spreadsheet"


@dataclasses.dataclass
class PolarsDatabaseReader(DataLoader):
    """
    Class specifically to handle loading DataFrame from a database.
    """

    query: str
    connection: str
    # kwargs:
    iter_batches: bool = False
    batch_size: Union[int, None] = None
    schema_overrides: Union[Dict[str, Any], None] = None
    infer_schema_length: Union[int, None] = None
    execute_options: Union[Dict[str, Any], None] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.iter_batches is not None:
            kwargs["iter_batches"] = self.iter_batches
        if self.batch_size is not None:
            kwargs["batch_size"] = self.batch_size
        if self.schema_overrides is not None:
            kwargs["schema_overrides"] = self.schema_overrides
        if self.infer_schema_length is not None:
            kwargs["infer_schema_length"] = self.infer_schema_length
        if self.execute_options is not None:
            kwargs["execute_options"] = self.execute_options
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.read_database(
            query=self.query,
            connection=self.connection,
            **self._get_loading_kwargs(),
        )
        metadata = utils.get_file_and_dataframe_metadata(self.query, df)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "database"


@dataclasses.dataclass
class PolarsDatabaseWriter(DataSaver):
    """
    Class specifically to handle saving DataFrame to a database.
    """

    table_name: str
    connection: str
    if_table_exists: Literal["fail", "replace", "append"] = "fail"
    engine: Literal["auto", "sqlalchemy", "adbc"] = "sqlalchemy"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE, pl.LazyFrame]

    def _get_saving_kwargs(self):
        kwargs = {}
        if self.if_table_exists is not None:
            kwargs["if_table_exists"] = self.if_table_exists
        if self.engine is not None:
            kwargs["engine"] = self.engine
        return kwargs

    def save_data(self, data: Union[DATAFRAME_TYPE, pl.LazyFrame]) -> Dict[str, Any]:
        if isinstance(data, pl.LazyFrame):
            data = data.collect()

        data.write_database(
            table_name=self.table_name,
            connection=self.connection,
            **self._get_saving_kwargs(),
        )
        return utils.get_file_and_dataframe_metadata(self.table_name, data)

    @classmethod
    def name(cls) -> str:
        return "database"


def register_data_loaders():
    """Function to register the data loaders for this extension."""
    for loader in [
        PolarsCSVReader,
        PolarsCSVWriter,
        PolarsParquetReader,
        PolarsParquetWriter,
        PolarsFeatherReader,
        PolarsFeatherWriter,
        PolarsAvroReader,
        PolarsAvroWriter,
        PolarsJSONReader,
        PolarsJSONWriter,
        PolarsDatabaseReader,
        PolarsDatabaseWriter,
        PolarsSpreadsheetReader,
        PolarsSpreadsheetWriter,
    ]:
        registry.register_adapter(loader)



import logging

from hamilton.plugins.h_dask import DaskDataFrameResult, DaskExecutor, DaskGraphAdapter

logger = logging.getLogger(__name__)
logger.warning(
    "Importing from this module is deprecated. We have moved these features out of experimental! "
    "Please use hamilton.plugins.h_dask instead."
)

__all__ = ["DaskGraphAdapter", "DaskDataFrameResult", "DaskExecutor"]



from hamilton.function_modifiers import config


def new_param() -> str:
    return "dummy"


@config.when(fn_1_version=0)
def fn() -> str:
    pass


@config.when(fn_1_version=1)
def fn__v1() -> str:
    return "version_1"


@config.when(fn_1_version=2)
def fn__v2(new_param: str) -> str:
    return "version_2"


@config.when(fn_1_version=3, name="fn")
def fn_to_rename() -> str:
    return "version_3"



from typing import Callable, Dict, List, Union

import pytest

from hamilton import node
from hamilton.execution.graph_functions import (
    create_input_string,
    nodes_between,
    topologically_sort_nodes,
)


def _create_dummy_dag(
    adjacency_map: Dict[str, List[str]], dict_output: bool = False
) -> Union[List[node.Node], Dict[str, node.Node]]:
    name_map = {}
    for name, dependencies in adjacency_map.items():
        input_types = {dep: object for dep in dependencies}
        node_ = node.Node(
            name=name,
            typ=object,
            callabl=lambda **kwargs: object(),
            input_types=input_types,
        )
        name_map[name] = node_
    nodes = []
    for name, dependencies in adjacency_map.items():
        node_ = name_map[name]
        for dependency in dependencies:
            dep = name_map[dependency]
            dep.depended_on_by.append(node_)
            node_.dependencies.append(name_map[dependency])
        nodes.append(node_)
    if dict_output:
        return {node_.name: node_ for node_ in nodes}
    return nodes


def _assert_topologically_sorted(nodes, sorted_nodes):
    for node_ in nodes:
        for dep in node_.dependencies:
            assert sorted_nodes.index(dep.name) < sorted_nodes.index(node_.name)


@pytest.mark.parametrize(
    "dag_input, expected_sorted_nodes",
    [
        ({"a": [], "b": ["a"], "c": ["a"], "d": ["b", "c"], "e": ["d"]}, ["a", "b", "c", "d", "e"]),
        ({}, []),
        ({"a": []}, ["a"]),
        (
            {
                "a": ["b", "c"],
                "b": ["d", "e"],
                "c": ["d", "e"],
                "d": ["f"],
                "e": ["f", "g"],
                "f": ["h"],
                "g": ["h", "i"],
                "h": ["j", "k"],
                "i": ["k", "l"],
                "j": ["m"],
                "k": ["m", "n"],
                "l": ["n"],
                "m": ["o"],
                "n": ["o", "p"],
                "o": ["q"],
                "p": ["q"],
                "q": [],
            },
            ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q"],
        ),
    ],
    ids=[
        "Simple DAG",
        "Empty DAG",
        "Single Node DAG",
        "Large DAG",
    ],
)
def test_topologically_sort_nodes(dag_input, expected_sorted_nodes):
    nodes = _create_dummy_dag(dag_input)
    sorted_nodes = [item.name for item in topologically_sort_nodes(nodes)]
    _assert_topologically_sorted(nodes, sorted_nodes)


def _is(name: str) -> Callable[[node.Node], bool]:
    def _inner(n: node.Node) -> bool:
        return n.name == name

    return _inner


@pytest.mark.parametrize(
    "dag_repr, expected_nodes_in_between, start_node, end_node",
    [
        (
            {"a": [], "b": ["a"], "c": ["b"]},
            {"b"},
            "a",
            "c",
        ),
        (
            {"a": [], "b": ["a"], "c": ["b"], "d": ["c"], "e": ["d"]},
            {"b", "c", "d"},
            "a",
            "e",
        ),
        (
            {
                "a": ["b", "c"],
                "b": ["d"],
                "c": ["d"],
                "d": ["e"],
                "e": ["f"],
                "f": ["g"],
                "g": ["h"],
                "h": ["i", "j"],
                "i": ["k"],
                "j": ["k"],
                "k": ["l"],
                "l": ["m", "n"],
                "m": ["o"],
                "n": ["o"],
                "o": ["p"],
                "p": ["q"],
                "q": [],
            },
            {"e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p"},
            "q",
            "d",
        ),
        ({"a": [], "b": [], "c": ["a", "b"], "d": "c"}, {"c"}, "a", "d"),
        # https://github.com/DAGWorks-Inc/hamilton/issues/948
        (
            {
                "random_int": [],
                "numbers": ["random_int"],
                "add1": ["numbers", "random_int"],
                "add2": ["add1", "random_int"],
                "collect_numbers": ["add2"],
                "final_result": ["collect_numbers"],
            },
            {"add1", "add2"},
            "numbers",
            "collect_numbers",
        ),
    ],
    ids=["simple_base", "longer_chain", "complex_dag", "subdag_with_external_dep", "issue_948"],
)
def test_find_nodes_between(dag_repr, expected_nodes_in_between, start_node, end_node):
    nodes = _create_dummy_dag(dag_repr, dict_output=True)
    found_node, in_between = nodes_between(nodes[end_node], _is(start_node))
    assert found_node.name == start_node
    assert set(in_between) == {nodes[item] for item in expected_nodes_in_between}


def test_create_input_string_with_short_values():
    """Tests that create_input_string works correctly with short values"""
    kwargs = {"arg1": 1, "arg2": "short string", "arg3": 3.14}
    result = create_input_string(kwargs)
    assert result == "{'arg1': 1, 'arg2': 'short string', 'arg3': 3.14}"


def test_create_input_string_with_long_values():
    """Tests that create_input_string truncates long values"""
    kwargs = {"arg1": "a" * 51, "arg2": "b" * 52, "arg3": "c" * 53}
    result = create_input_string(kwargs)
    assert result == (
        "{'arg1': \"'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...\",\n"
        " 'arg2': \"'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb...\",\n"
        " 'arg3': \"'ccccccccccccccccccccccccccccccccccccccccccccccccc...\"}"
    )


def test_create_input_string_with_empty_dict():
    """Tests that create_input_string works correctly with an empty dictionary"""
    kwargs = {}
    result = create_input_string(kwargs)
    assert result == "{}"


def test_create_input_string_with_large_number_of_args():
    """Tests that create_input_string truncates the output if there are too many arguments"""
    kwargs = {f"arg{i}": i for i in range(1000)}
    result = create_input_string(kwargs)
    assert result.startswith("{")
    assert result.endswith("...")
    assert len(result) == 1003


def test_create_input_string_with_dataframes():
    """Tests that create_input_string works correctly with pandas dataframes"""
    import pandas as pd

    kwargs = {
        "arg1": pd.DataFrame({"a": [1, 2, 3] * 10, "b": [4, 5, 6] * 10}),
        "arg2": "short string",
        "arg3": 3.14,
    }
    result = create_input_string(kwargs)
    assert result == (
        "{'arg1': '    a  b\\n0   1  4\\n1   2  5\\n2   3  6\\n3   1  4\\n4   2...',\n"
        " 'arg2': 'short string',\n"
        " 'arg3': 3.14}"
    )


def test_create_input_string_with_custom_object():
    """Tests that create_input_string works correctly with custom objects"""

    class CustomObject:
        def __init__(self, a, b):
            self.a = a
            self.b = b

    kwargs = {"arg1": CustomObject(1, 2), "arg2": "short string", "arg3": 3.14}
    result = create_input_string(kwargs)
    assert result == (
        "{'arg1': '<tests.execution.test_graph_functions.test_create_...',\n"
        " 'arg2': 'short string',\n"
        " 'arg3': 3.14}"
    )



import logging
from typing import Any, Type

from packaging import version

from hamilton import registry

logger = logging.getLogger(__name__)

try:
    from xlsxwriter.workbook import Workbook
except ImportError:
    Workbook = Type

try:
    import polars as pl
except ImportError:
    raise NotImplementedError("Polars is not installed.")

pl_version = version.Version(pl.__version__)
if pl_version < version.Version("1.0.0"):
    from hamilton.plugins.polars_pre_1_0_0_extension import register_data_loaders

    logger.warning(
        "Using pre-1.0.0 Polars integration -- we will stop supporting this in Hamilton 2.0, so please upgrade your version of polars! "
        f"Current version: {pl_version}, minimum required version: 1.0.0."
    )
else:
    from hamilton.plugins.polars_post_1_0_0_extension import register_data_loaders

register_data_loaders()

DATAFRAME_TYPE = pl.DataFrame
COLUMN_TYPE = pl.Series


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("polars", DATAFRAME_TYPE, COLUMN_TYPE)


@registry.get_column.register(pl.DataFrame)
def get_column_polars(df: pl.DataFrame, column_name: str) -> pl.Series:
    return df[column_name]


@registry.fill_with_scalar.register(pl.DataFrame)
def fill_with_scalar_polars(df: pl.DataFrame, column_name: str, scalar_value: Any) -> pl.DataFrame:
    if not isinstance(scalar_value, pl.Series):
        scalar_value = [scalar_value]
    return df.with_column(pl.Series(name=column_name, values=scalar_value))


register_types()



"""
This module should not have any real business logic.

It should only house the graph & things required to create and traverse one.

Note: one should largely consider the code in this module to be "private".
"""

import inspect
import logging
import os.path
import pathlib
import uuid
from enum import Enum
from types import ModuleType
from typing import Any, Callable, Collection, Dict, FrozenSet, List, Optional, Set, Tuple, Type

import hamilton.lifecycle.base as lifecycle_base
from hamilton import graph_types, node
from hamilton.execution import graph_functions
from hamilton.function_modifiers import base as fm_base
from hamilton.function_modifiers.metadata import schema
from hamilton.graph_utils import find_functions
from hamilton.htypes import get_type_as_string, types_match
from hamilton.node import Node

logger = logging.getLogger(__name__)


class VisualizationNodeModifiers(Enum):
    """Enum of all possible node modifiers for visualization."""

    IS_OUTPUT = 1
    IS_PATH = 2
    IS_USER_INPUT = 3
    IS_OVERRIDE = 4


def add_dependency(
    func_node: node.Node,
    func_name: str,
    nodes: Dict[str, node.Node],
    param_name: str,
    param_type: Type,
    adapter: lifecycle_base.LifecycleAdapterSet,
):
    """Adds dependencies to the node objects.

    This will add user defined inputs to the dictionary of nodes in the graph.

    :param func_node: the node we're pulling dependencies from.
    :param func_name: the name of the function we're inspecting.
    :param nodes: nodes representing the graph. This function mutates this object and underlying objects.
    :param param_name: the parameter name we're looking for/adding as a dependency.
    :param param_type: the type of the parameter.
    :param adapter: The adapter that adapts our node type checking based on the context.
    """
    adapter_checks_types = adapter.does_method("do_check_edge_types_match", is_async=False)
    if param_name in nodes:
        # validate types match
        required_node: Node = nodes[param_name]
        types_do_match = types_match(param_type, required_node.type)
        types_do_match |= adapter_checks_types and adapter.call_lifecycle_method_sync(
            "do_check_edge_types_match",
            type_from=param_type,
            type_to=required_node.type,
        )
        if not types_do_match and required_node.user_defined:
            # check the case that two input type expectations are compatible, e.g. is one a subset of the other
            # this could be the case when one is a union and the other is a subset of that union
            # which is fine for inputs. If they are not compatible, we raise an error.
            types_are_compatible = types_match(required_node.type, param_type)
            types_are_compatible |= adapter_checks_types and adapter.call_lifecycle_method_sync(
                "do_check_edge_types_match",
                type_from=param_type,
                type_to=required_node.type,
            )
            if not types_are_compatible:
                raise ValueError(
                    f"Error: Two or more functions are requesting {param_name}, but have incompatible types. "
                    f"{func_name} requires {param_name} to be {param_type}, but found another function(s) "
                    f"{[f.__name__ for f in required_node.originating_functions]} that minimally require {param_name} "
                    f"as {required_node.type}. Please fix this by ensuring that all functions requesting {param_name} "
                    f"have compatible types. If you believe they are equivalent, please reach out to the developers. "
                    f"Note that, if you have types that are equivalent for your purposes, you can create a "
                    f"graph adapter that checks the types against each other in a more lenient manner."
                )
            else:
                # replace the current type with this "tighter" type.
                required_node.set_type(param_type)
                # add to the originating functions
                for og_func in func_node.originating_functions:
                    required_node.add_originating_function(og_func)
        elif not types_do_match:
            raise ValueError(
                f"Error: {func_name} is expecting {param_name}:{param_type}, but found "
                f"{param_name}:{required_node.type}. \nHamilton does not consider these types to be "
                f"equivalent. If you believe they are equivalent, please reach out to the developers. "
                f"Note that, if you have types that are equivalent for your purposes, you can create a "
                f"graph adapter that checks the types against each other in a more lenient manner."
            )
    else:
        # this is a user defined var, i.e. an input to the graph.
        required_node = node.Node(
            param_name,
            param_type,
            node_source=node.NodeType.EXTERNAL,
            originating_functions=func_node.originating_functions,
        )
        nodes[param_name] = required_node
    # add edges
    func_node.dependencies.append(required_node)
    required_node.depended_on_by.append(func_node)


def update_dependencies(
    nodes: Dict[str, node.Node],
    adapter: lifecycle_base.LifecycleAdapterSet,
    reset_dependencies: bool = True,
):
    """Adds dependencies to a dictionary of nodes. If in_place is False,
    it will deepcopy the dict + nodes and return that. Otherwise it will
    mutate + return the passed-in dict + nodes.

    Note: this will add in "input" nodes if they are not already present.

    :param in_place: Whether or not to modify in-place, or copy/return
    :param nodes: Nodes that form the DAG we're updating
    :param adapter: Adapter to use for type checking
    :param reset_dependencies: Whether or not to reset the dependencies. If they are not set this is
    unnecessary, and we can save yet another pass. Note that `reset` will perform an in-place
    operation.
    :return: The updated nodes
    """
    # copy without the dependencies to avoid duplicates
    if reset_dependencies:
        nodes = {k: v.copy(include_refs=False) for k, v in nodes.items()}
    for node_name, n in list(nodes.items()):
        for param_name, (param_type, _) in n.input_types.items():
            add_dependency(n, node_name, nodes, param_name, param_type, adapter)
    return nodes


def create_function_graph(
    *modules: ModuleType,
    config: Dict[str, Any],
    adapter: lifecycle_base.LifecycleAdapterSet = None,
    fg: Optional["FunctionGraph"] = None,
) -> Dict[str, node.Node]:
    """Creates a graph of all available functions & their dependencies.
    :param modules: A set of modules over which one wants to compute the function graph
    :param config: Dictionary that we will inspect to get values from in building the function graph.
    :param adapter: The adapter that adapts our node type checking based on the context.
    :return: list of nodes in the graph.
    If it needs to be more complicated, we'll return an actual networkx graph and get all the rest of the logic for free
    """
    if adapter is None:
        adapter = (
            lifecycle_base.LifecycleAdapterSet()
        )  # empty one -- not provided/necessary, we can run without it
    if fg is None:
        nodes = {}  # name -> Node
    else:
        nodes = fg.nodes
    functions = sum([find_functions(module) for module in modules], [])

    # create non-input nodes -- easier to just create this in one loop
    for func_name, f in functions:
        for n in fm_base.resolve_nodes(f, config):
            if n.name in config:
                continue  # This makes sure we overwrite things if they're in the config...
            if n.name in nodes:
                raise ValueError(
                    f"Cannot define function {n.name} more than once."
                    f" Already defined by function {f}"
                )
            nodes[n.name] = n
    # add dependencies -- now that all nodes except input nodes, we just run through edges & validate graph.
    nodes = update_dependencies(nodes, adapter, reset_dependencies=False)  # no dependencies
    # present yet
    for key in config.keys():
        if key not in nodes:
            nodes[key] = node.Node(key, Any, node_source=node.NodeType.EXTERNAL)
    return nodes


def _check_keyword_args_only(func: Callable) -> bool:
    """Checks if a function only takes keyword arguments."""
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if param.default is inspect.Parameter.empty and param.kind not in [
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.VAR_KEYWORD,
        ]:
            return False
    return True


def create_graphviz_graph(
    nodes: Set[node.Node],
    comment: str,
    graphviz_kwargs: dict,
    node_modifiers: Dict[str, Set[VisualizationNodeModifiers]],
    strictly_display_only_nodes_passed_in: bool,
    show_legend: bool = True,
    orient: str = "LR",
    hide_inputs: bool = False,
    deduplicate_inputs: bool = False,
    display_fields: bool = True,
    custom_style_function: Callable = None,
    config: dict = None,
) -> "graphviz.Digraph":  # noqa: F821
    """Helper function to create a graphviz graph.

    :param nodes: The set of computational nodes
    :param comment: The comment to have on the graph.
    :param graphviz_kwargs: kwargs to pass to create the graph.
        e.g. dict(graph_attr={'ratio': '1'}) will set the aspect ratio to be equal of the produced image.
    :param node_modifiers: A dictionary of node names to dictionaries of node attributes to modify.
    :param strictly_display_only_nodes_passed_in: If True, only display the nodes passed in. Else defaults to displaying
        also what nodes a node depends on (i.e. all nodes that feed into it).
    :param show_legend: If True, add a legend to the visualization based on the DAG's nodes.
    :param orient: `LR` stands for "left to right". Accepted values are TB, LR, BT, RL.
        `orient` will be overwridden by the value of `graphviz_kwargs['graph_attr']['rankdir']`
        see (https://graphviz.org/docs/attr-types/rankdir/)
    :param hide_inputs: If True, no input nodes are displayed.
    :param deduplicate_inputs: If True, remove duplicate input nodes.
        Can improve readability depending on the specifics of the DAG.
    :param custom_style_function: A function that takes in node values and returns a dictionary of styles to apply to it.
    :param config: The Driver config. This value is passed by the caller, e.g., driver.display_all_functions(),
        and shouldn't be passed explicitly. Otherwise, it may not match the `nodes` argument leading to an
        incorrect visualization.
    :return: a graphviz.Digraph; use this to render/save a graph representation.
    """
    PATH_COLOR = "red"

    import graphviz

    if custom_style_function is not None:
        if not _check_keyword_args_only(custom_style_function):
            raise ValueError(
                "custom_style_function must only take keyword arguments"
                " `node` and `node_class`. E.g. Signature should resemble:\n"
                "def custom_style(*,  # <--- key part is this * \n"
                "     node: graph_types.HamiltonNode, \n"
                "     node_class: str\n"
                ") -> Tuple[dict, Optional[str], Optional[str]]:"
            )

    if config is None:
        raise ValueError(
            "Received None for kwarg `config`. Make sure to pass a dictionary that matches the Driver config.\n"
            "If you're seeing this error, you're likely using a non-public API."
        )

    def _get_node_label(
        n: node.Node,
        name: Optional[str] = None,
        type_string: Optional[str] = None,
    ) -> str:
        """Get a graphviz HTML-like node label. It uses the DAG node
        name and type but values can be overridden. Overriding is currently
        used for materializers since `type_` is stored in n.tags.

        ref: https://graphviz.org/doc/info/shapes.html#html
        """
        name = n.name if name is None else name
        if type_string is None:
            type_string = get_type_as_string(n.type) if get_type_as_string(n.type) else ""

        return f"<<b>{name}</b><br /><br /><i>{type_string}</i>>"

    def _get_input_label(input_nodes: FrozenSet[node.Node]) -> str:
        """Get a graphviz HTML-like node label formatted aspyer a table.
        Each row is a different input node with one column containing
        the name and the other the type.
        ref: https://graphviz.org/doc/info/shapes.html#html
        """
        rows = []
        for dep in input_nodes:
            name = dep.name
            type_string = get_type_as_string(dep.type) if get_type_as_string(dep.type) else ""
            rows.append(f"<tr><td>{name}</td><td>{type_string}</td></tr>")
        return f"<<table border=\"0\">{''.join(rows)}</table>>"

    def _get_node_type(n: node.Node) -> str:
        """Get the node type of a DAG node.

        Input: is external, doesn't originate from a function, functions depend on it
        Config: is external, doesn't originate from a function, no function depedends on it
        Function: others
        """
        if n._node_source == node.NodeType.EXTERNAL and n._depended_on_by:
            return "input"
        elif (
            n._node_source == node.NodeType.EXTERNAL
            and n._originating_functions is None
            and not n._depended_on_by
        ):
            return "config"
        else:
            return "function"

    def _get_node_style(node_type: str) -> Dict[str, str]:
        """Get the style of a node type.
        Graphviz needs values to be strings.
        """
        fontname = "Helvetica"

        if node_type == "config":
            node_style = dict(
                shape="note",
                style="filled",
                fontname=fontname,
            )
        elif node_type == "input":
            node_style = dict(
                shape="rectangle",
                margin="0.15",
                style="filled,dashed",
                fontname=fontname,
            )
        elif node_type == "materializer":
            node_style = dict(
                shape="cylinder",
                style="filled",
                margin="0.15,0.1",
                fontname=fontname,
            )
        else:  # this is a function or else
            node_style = dict(
                shape="rectangle",
                margin="0.15",
                style="rounded,filled",
                fillcolor="#b4d8e4",
                fontname=fontname,
            )

        return node_style

    def _get_function_modifier_style(modifier: str) -> Dict[str, str]:
        """Get the style of a modifier. The dictionary returned
        is used to overwrite values of the base node style.
        Graphviz needs values to be strings.
        """
        if modifier == "output":
            modifier_style = dict(fillcolor="#FFC857")
        elif modifier == "collect":
            modifier_style = dict(peripheries="2", color="#EA5556")
        elif modifier == "expand":
            modifier_style = dict(peripheries="2", color="#56E39F")
        elif modifier == "override":
            modifier_style = dict(style="filled,diagonals")
        elif modifier == "materializer":
            modifier_style = dict(shape="cylinder")
        elif modifier == "field":
            modifier_style = dict(fillcolor="#c8dae0", fontname="Courier")
        elif modifier == "cluster":
            modifier_style = dict(
                fillcolor="#ffffff", color="#649fb3", style="rounded,filled,dashed"
            )
        else:
            modifier_style = dict()

        return modifier_style

    def _get_edge_style(from_type: str, to_type: str) -> Dict:
        """

        Graphviz needs values to be strings.
        """
        edge_style = dict()

        if from_type == "expand":
            edge_style.update(
                dir="both",
                arrowhead="crow",
                arrowtail="none",
            )

        if to_type == "collect":
            edge_style.update(dir="both", arrowtail="crow")

        return edge_style

    def _get_legend(
        node_types: Set[str], extra_legend_nodes: Dict[Tuple[str, str], Dict[str, str]]
    ):
        """Create a visualization legend as a graphviz subgraph. The legend includes the
        node types and modifiers presente in the visualization.
        """
        legend_subgraph = graphviz.Digraph(
            name="cluster__legend",  # needs to start with `cluster` for graphviz layout
            graph_attr=dict(
                label="Legend",
                rank="same",  # makes the legend perpendicular to the main DAG
                fontname="helvetica",
                fillcolor="#ffffff",
            ),
        )

        sorted_types = [
            "config",
            "input",
            "function",
            "cluster",
            "field",
            "output",
            "materializer",
            "override",
            "expand",
            "collect",
        ]

        for node_type in sorted(node_types, key=lambda t: sorted_types.index(t)):
            node_style = _get_node_style(node_type)
            modifier_style = _get_function_modifier_style(node_type)
            node_style.update(**modifier_style)
            legend_subgraph.node(name=node_type, **node_style)

        for (base_class, node_name), legend_style in extra_legend_nodes.items():
            base_style = _get_node_style(base_class)
            base_style.update(**legend_style)
            legend_subgraph.node(name=node_name, **base_style)

        return legend_subgraph

    # handle default values in nested dict
    digraph_attr = dict(
        comment=comment,
        graph_attr=dict(
            rankdir=orient,
            ranksep="0.4",
            compound="true",
            concentrate="true",
            style="filled",
            # bgcolor="transparent",
        ),
        node_attr=dict(
            fillcolor="#ffffff",
        ),
        edge_attr=dict(
            # color="#ffffff",
        ),
    )
    # we need to update the graph_attr dict instead of overwriting it
    # so that means we need to handle nested dicts, e.g. graph_attr.
    for g_key, g_value in graphviz_kwargs.items():
        if isinstance(g_value, dict):
            digraph_attr[g_key].update(**g_value)
        else:
            digraph_attr[g_key] = g_value

    digraph = graphviz.Digraph(**digraph_attr)
    extra_legend_nodes = {}

    for config_key, config_value in config.items():
        label = _get_node_label(n=None, name=config_key, type_string=str(config_value))
        style = _get_node_style("config")
        digraph.node(config_key, label=label, **style)

    # create nodes
    seen_node_types = set()
    for n in nodes:
        label = _get_node_label(n)
        node_type = _get_node_type(n)
        if node_type == "input":
            seen_node_types.add(node_type)
            continue
        # config nodes are handled separately;
        # only Driver.display_all_functions() passes config via the `nodes` arg
        elif node_type == "config":
            continue

        # append config key to node label
        config_key = n.tags.get("hamilton.config", None)
        if config_key:
            seen_node_types.add("config")
            label = _get_node_label(n, name=f"{n.name}: {config_key}")

        node_style = _get_node_style(node_type)

        # prefer having the conditions explicit for now since they rely on
        # heterogeneous VisualizationNodeModifiers and node.Node.node_role.
        # Otherwise, it's difficult to manage seen nodes and the legend.
        if n.node_role == node.NodeType.EXPAND:
            modifier_style = _get_function_modifier_style("expand")
            node_style.update(**modifier_style)
            seen_node_types.add("expand")

        if n.node_role == node.NodeType.COLLECT:
            modifier_style = _get_function_modifier_style("collect")
            node_style.update(**modifier_style)
            seen_node_types.add("collect")

        if n.tags.get("hamilton.data_saver"):
            materializer_type = n.tags["hamilton.data_saver.classname"]
            label = _get_node_label(n, type_string=materializer_type)
            modifier_style = _get_function_modifier_style("materializer")
            node_style.update(**modifier_style)
            seen_node_types.add("materializer")

        if n.tags.get("hamilton.data_loader") and "load_data." in n.name:
            materializer_type = n.tags["hamilton.data_loader.classname"]
            label = _get_node_label(n, type_string=materializer_type)
            modifier_style = _get_function_modifier_style("materializer")
            node_style.update(**modifier_style)
            seen_node_types.add("materializer")

        # apply custom styles before node modifiers
        seen_node_type = None
        if custom_style_function:
            custom_style, base_type, legend_name = custom_style_function(
                node=graph_types.HamiltonNode.from_node(n), node_class=node_type
            )
            if legend_name:
                extra_legend_nodes[(base_type, legend_name)] = custom_style
                seen_node_type = legend_name
            node_style.update(**custom_style)

        if node_modifiers.get(n.name):
            modifiers = node_modifiers[n.name]
            if VisualizationNodeModifiers.IS_OUTPUT in modifiers:
                modifier_style = _get_function_modifier_style("output")
                node_style.update(**modifier_style)
                seen_node_types.add("output")

            if VisualizationNodeModifiers.IS_OVERRIDE in modifiers:
                modifier_style = _get_function_modifier_style("override")
                node_style.update(**modifier_style)
                seen_node_types.add("override")

            if VisualizationNodeModifiers.IS_PATH in modifiers:
                # use PATH_COLOR only if no color applied, edges provide enough clarity
                # currently, only EXPAND and COLLECT use the `color` attribue
                node_style["color"] = node_style.get("color", PATH_COLOR)

        digraph.node(n.name, label=label, **node_style)

        # only do field-level visualization if there's a schema specified and we want to display it
        if n.tags.get(schema.INTERNAL_SCHEMA_OUTPUT_KEY) and display_fields:
            # When a node has attached schema data -> add a cluster with a node for each field
            seen_node_types.add("cluster")
            seen_node_types.add("field")

            def _create_equal_length_cols(schema_tag: str) -> List[str]:
                cols = schema_tag.split(",")
                for i in range(len(cols)):

                    def _insert_space_after_colon(col: str) -> str:
                        colon_index = col.find(":")
                        if colon_index != -1 and col[colon_index + 1] != " ":
                            return col[: colon_index + 1] + " " + col[colon_index + 1 :]
                        return col

                    cols[i] = _insert_space_after_colon(cols[i].strip())
                max_length = max([len(col) for col in cols])
                return [col.ljust(max_length) for col in cols]

            cluster_node_style = node_style.copy()
            cluster_node_style.update(
                {
                    "color": "#649fb3",
                    "style": "rounded,filled,dashed",
                    "fillcolor": "#ffffff",
                    "margin": "10",
                }
            )
            field_node_style = node_style.copy()
            field_node_style.update(
                {
                    "fillcolor": "#c8dae0",
                    "fontname": "Courier",
                    "fontsize": "10",
                }
            )
            with digraph.subgraph(name="cluster_" + n.name) as c:
                c.attr(**cluster_node_style)
                cols = _create_equal_length_cols(n.tags.get(schema.INTERNAL_SCHEMA_OUTPUT_KEY))
                for i in range(len(cols)):
                    # We may want to change the name. We have a different name than label,
                    # as it crowds out the global namespace
                    # For now, however, this should be unique
                    c.node(n.name + ":" + cols[i], **field_node_style, label=cols[i])
                c.node(n.name)

        if seen_node_type is None:
            seen_node_types.add(node_type)

    # create edges
    input_sets = dict()
    for n in nodes:
        to_type = "collect" if n.node_role == node.NodeType.COLLECT else ""
        to_modifiers = node_modifiers.get(n.name, set())

        input_nodes = set()
        for d in n.dependencies:
            if strictly_display_only_nodes_passed_in and d not in nodes:
                continue

            dependency_type = _get_node_type(d)
            # input nodes and edges are gathered instead of drawn
            # they are drawn later, see below
            if dependency_type == "input":
                input_nodes.add(d)
                continue

            from_type = "expand" if d.node_role == node.NodeType.EXPAND else ""
            dependency_modifiers = node_modifiers.get(d.name, set())
            edge_style = _get_edge_style(from_type, to_type)
            if (
                VisualizationNodeModifiers.IS_PATH in dependency_modifiers
                and VisualizationNodeModifiers.IS_PATH in to_modifiers
            ):
                edge_style["color"] = PATH_COLOR

            digraph.edge(d.name, n.name, **edge_style)

        # skip input node creation
        if hide_inputs:
            continue

        # draw input nodes if at least 1 exist
        if len(input_nodes) > 0:
            input_node_name = f"_{n.name}_inputs"

            # following block is for input node deduplication
            input_nodes = frozenset(input_nodes)
            if input_sets.get(input_nodes):
                existing_input_name = input_sets[input_nodes]
                digraph.edge(existing_input_name, n.name)
                continue

            # allow duplicate input nodes by never storing keys
            if deduplicate_inputs:
                input_sets[input_nodes] = input_node_name

            # create input node
            node_label = _get_input_label(input_nodes)
            node_style = _get_node_style("input")
            digraph.node(name=input_node_name, label=node_label, **node_style)
            # create edge for input node
            digraph.edge(input_node_name, n.name)

    if show_legend:
        digraph.subgraph(_get_legend(seen_node_types, extra_legend_nodes))

    return digraph


def create_networkx_graph(
    nodes: Set[node.Node], user_nodes: Set[node.Node], name: str
) -> "networkx.DiGraph":  # noqa: F821
    """Helper function to create a networkx graph.

    :param nodes: The set of computational nodes
    :param user_nodes: The set of nodes that the user is providing inputs for.
    :param name: The name to have on the graph.
    :return: a graphviz.Digraph; use this to render/save a graph representation.
    """
    import networkx

    digraph = networkx.DiGraph(name=name)
    for n in nodes:
        digraph.add_node(n.name, label=n.name)
    for n in user_nodes:
        digraph.add_node(n.name, label=f"UD: {n.name}")

    for n in list(nodes) + list(user_nodes):
        for d in n.dependencies:
            digraph.add_edge(d.name, n.name)
    return digraph


class FunctionGraph:
    """Note: this object should be considered private until stated otherwise.

    That is, you should not try to build off of it directly without chatting to us first.
    """

    def __init__(
        self,
        nodes: Dict[str, Node],
        config: Dict[str, Any],
        adapter: lifecycle_base.LifecycleAdapterSet = None,
    ):
        """Initializes a function graph from specified nodes. See note on `from_modules` if you
        start getting an error here because you use an internal API.

        :param nodes: Nodes, taken from the output of create_function_graph.
        :param config: this is configuration and/or initial data.
        :param adapter: Group of adapters, to use for node resolution.
        """
        if adapter is None:
            adapter = lifecycle_base.LifecycleAdapterSet()

        self._config = config
        self.nodes = nodes
        self.adapter = adapter

    @staticmethod
    def from_modules(
        *modules: ModuleType,
        config: Dict[str, Any],
        adapter: lifecycle_base.LifecycleAdapterSet = None,
    ):
        """Initializes a function graph from the specified modules. Note that this was the old
        way we constructed FunctionGraph -- this is not a public-facing API, so we replaced it
        with a constructor that takes in nodes directly. If you hacked in something using
        `FunctionGraph`, then you should be able to replace it with `FunctionGraph.from_modules`
        and it will work.

        :param modules: Modules to crawl, resolve to nodes
        :param config: Config to use for node resolution
        :param adapter: All adapters, to use for node resolution
        :return: a function graph.
        """

        nodes = create_function_graph(*modules, config=config, adapter=adapter)
        return FunctionGraph(nodes, config, adapter)

    def with_nodes(self, nodes: Dict[str, Node]) -> "FunctionGraph":
        """Creates a new function graph with the additional specified nodes.
        Note that if there is a duplication in the node definitions,
        it will error out.

        :param nodes: Nodes to add to the FunctionGraph
        :return: a new function graph.
        """
        # first check if there are any duplicates
        duplicates = set(self.nodes.keys()).intersection(set(nodes.keys()))
        if duplicates:
            raise ValueError(f"Duplicate node names found: {duplicates}")
        new_node_dict = update_dependencies({**self.nodes, **nodes}, self.adapter)
        return FunctionGraph(new_node_dict, self.config, self.adapter)

    @property
    def config(self):
        return self._config

    @property
    def decorator_counter(self) -> Dict[str, int]:
        return fm_base.DECORATOR_COUNTER

    def get_nodes(self) -> List[node.Node]:
        return list(self.nodes.values())

    def display_all(
        self,
        output_file_path: str = "test-output/graph-all.gv",
        render_kwargs: dict = None,
        graphviz_kwargs: dict = None,
        show_legend: bool = True,
        orient: str = "LR",
        hide_inputs: bool = False,
        deduplicate_inputs: bool = False,
        display_fields: bool = True,
        custom_style_function: Callable = None,
        keep_dot: bool = False,
    ) -> Optional["graphviz.Digraph"]:  # noqa F821
        """Displays & saves a dot file of the entire DAG structure constructed.

        :param output_file_path: the place to save the files.
        :param render_kwargs: a dictionary of values we'll pass to graphviz render function. Defaults to viewing.
            If you do not want to view the file, pass in `{'view':False}`.
        :param graphviz_kwargs: kwargs to be passed to the graphviz graph object to configure it.
            e.g. dict(graph_attr={'ratio': '1'}) will set the aspect ratio to be equal of the produced image.
        :param show_legend: If True, add a legend to the visualization based on the DAG's nodes.
        :param orient: `LR` stands for "left to right". Accepted values are TB, LR, BT, RL.
            `orient` will be overwridden by the value of `graphviz_kwargs['graph_attr']['rankdir']`
            see (https://graphviz.org/docs/attr-types/rankdir/)
        :param hide_inputs: If True, no input nodes are displayed.
        :param deduplicate_inputs: If True, remove duplicate input nodes.
            Can improve readability depending on the specifics of the DAG.
        :param display_fields: If True, display fields in the graph if node has attached schema metadata
        :param custom_style_function: Optional. Custom style function.
        :param keep_dot: If true, produce a DOT file (ref: https://graphviz.org/doc/info/lang.html)
        :return: the graphviz graph object if it was created. None if not.
        """
        all_nodes = set()
        node_modifiers = {}
        for n in self.nodes.values():
            if n.user_defined:
                node_modifiers[n.name] = {VisualizationNodeModifiers.IS_USER_INPUT}
            all_nodes.add(n)
        if render_kwargs is None:
            render_kwargs = {}
        if graphviz_kwargs is None:
            graphviz_kwargs = {}
        return self.display(
            all_nodes,
            output_file_path=output_file_path,
            render_kwargs=render_kwargs,
            graphviz_kwargs=graphviz_kwargs,
            node_modifiers=node_modifiers,
            show_legend=show_legend,
            orient=orient,
            hide_inputs=hide_inputs,
            deduplicate_inputs=deduplicate_inputs,
            display_fields=display_fields,
            custom_style_function=custom_style_function,
            config=self._config,
            keep_dot=keep_dot,
        )

    def has_cycles(self, nodes: Set[node.Node], user_nodes: Set[node.Node]) -> bool:
        """Checks that the graph created does not contain cycles.

        :param nodes: the set of nodes that need to be computed.
        :param user_nodes: the set of inputs that the user provided.
        :return: bool. True if cycles detected. False if not.
        """
        cycles = self.get_cycles(nodes, user_nodes)
        return True if cycles else False

    def get_cycles(self, nodes: Set[node.Node], user_nodes: Set[node.Node]) -> List[List[str]]:
        """Returns cycles found in the graph.

        :param nodes: the set of nodes that need to be computed.
        :param user_nodes: the set of inputs that the user provided.
        :return: list of cycles, which is a list of node names.
        """
        try:
            import networkx
        except ModuleNotFoundError:
            logger.exception(
                " networkx is required for detecting cycles in the function graph. Install it with:"
                '\n\n  pip install "sf-hamilton[visualization]" or pip install networkx \n\n'
            )
            return []
        digraph = create_networkx_graph(nodes, user_nodes, "Dependency Graph")
        cycles = list(networkx.simple_cycles(digraph))
        return cycles

    @staticmethod
    def display(
        nodes: Set[node.Node],
        output_file_path: Optional[str] = None,
        render_kwargs: dict = None,
        graphviz_kwargs: dict = None,
        node_modifiers: Dict[str, Set[VisualizationNodeModifiers]] = None,
        strictly_display_only_passed_in_nodes: bool = False,
        show_legend: bool = True,
        orient: str = "LR",
        hide_inputs: bool = False,
        deduplicate_inputs: bool = False,
        display_fields: bool = True,
        custom_style_function: Callable = None,
        config: dict = None,
        keep_dot: bool = False,
    ) -> Optional["graphviz.Digraph"]:  # noqa F821
        """Function to display the graph represented by the passed in nodes.

        The output file format is determined through the following steps, each one overwriting the previous one:
        1. if `output_file_path` has no file extension, a PNG file is generated (e.g., `/path/to/file -> /path/to/file.png`)
        2. if `output_file_path` has a file extension, graphviz will use the specified format (e.g., `/path/to/file.svg -> /path/to/file.svg`))
        3. if a format value is specified for `render_kwargs={"format": "pdf"}`, it overrides any other inputs (e.g., /path/to/file.svg -> /path/to/file.pdf)

        :param nodes: the set of nodes that need to be computed.
        :param output_file_path: the path where we want to store the `dot` file + png picture. Pass in None to not save.
        :param render_kwargs: kwargs to be passed to the render function to visualize.
        :param graphviz_kwargs: kwargs to be passed to the graphviz graph object to configure it.
            e.g. dict(graph_attr={'ratio': '1'}) will set the aspect ratio to be equal of the produced image.
        :param node_modifiers: a dictionary of node names to a dictionary of attributes to modify.
            e.g. {'node_name': {NodeModifiers.IS_USER_INPUT}} will set the node named 'node_name' to be a user input.
        :param strictly_display_only_passed_in_nodes: if True, only display the nodes passed in.  Else defaults to
            displaying also what nodes a node depends on (i.e. all nodes that feed into it).
        :param show_legend: If True, add a legend to the visualization based on the DAG's nodes.
        :param orient: `LR` stands for "left to right". Accepted values are TB, LR, BT, RL.
            `orient` will be overridden by the value of `graphviz_kwargs['graph_attr']['rankdir']`
            see (https://graphviz.org/docs/attr-types/rankdir/)
        :param hide_inputs: If True, no input nodes are displayed.
        :param deduplicate_inputs: If True, remove duplicate input nodes.
            Can improve readability depending on the specifics of the DAG.
        :param display_fields: If True, display fields in the graph if node has attached
            schema metadata
        :param custom_style_function: Optional. Custom style function.
        :param config: The Driver config. This value is passed by the caller, e.g., driver.display_all_functions(),
            and shouldn't be passed explicitly. Otherwise, it may not match the `nodes` argument leading to an
            incorrect visualization.
        :return: the graphviz graph object if it was created. None if not.
        """
        # Check to see if optional dependencies have been installed.
        try:
            import graphviz  # noqa: F401
        except ModuleNotFoundError:
            logger.exception(
                " graphviz is required for visualizing the function graph. Install it with:"
                '\n\n  pip install "sf-hamilton[visualization]" or pip install graphviz \n\n'
            )
            return
        if graphviz_kwargs is None:
            graphviz_kwargs = {}
        if node_modifiers is None:
            node_modifiers = {}
        dot = create_graphviz_graph(
            nodes,
            "Dependency Graph",
            graphviz_kwargs,
            node_modifiers,
            strictly_display_only_passed_in_nodes,
            show_legend,
            orient,
            hide_inputs,
            deduplicate_inputs,
            display_fields=display_fields,
            custom_style_function=custom_style_function,
            config=config,
        )
        kwargs = {"format": "png"}  # default format = png
        if output_file_path:  # infer format from path
            output_file_path, suffix = os.path.splitext(output_file_path)
            if suffix != "":
                inferred_format = suffix.partition(".")[-1]
                kwargs.update(format=inferred_format)
        if render_kwargs and isinstance(render_kwargs, dict):  # accept explicit format
            kwargs.update(render_kwargs)
        # .render()` and `.pipe()` have quirks to handle separately
        # - `render()` accepts a `view` kwarg
        # - `render()` will append it's kwarg `format` to the filename
        if output_file_path:
            if keep_dot:
                kwargs["view"] = kwargs.get("view", False)
                dot.render(output_file_path, **kwargs)
            else:
                kwargs.pop("view", None)
                output_file_path = f"{output_file_path}.{kwargs['format']}"
                pathlib.Path(output_file_path).write_bytes(dot.pipe(**kwargs))
        return dot

    def get_impacted_nodes(self, var_changes: List[str]) -> Set[node.Node]:
        """DEPRECATED - use `get_downstream_nodes` instead."""
        logger.warning(
            "FunctionGraph.get_impacted_nodes is deprecated. "
            "Use `get_downstream_nodes` instead. This function will be removed"
            "in a future release."
        )
        return self.get_downstream_nodes(var_changes)

    def get_downstream_nodes(self, var_changes: List[str]) -> Set[node.Node]:
        """Given our function graph, and a list of nodes that are changed,
        returns the subgraph that they will impact.

        :param var_changes: the list of nodes that will change.
        :return: A set of all changed nodes.
        """
        nodes, user_nodes = self.directional_dfs_traverse(
            lambda n: n.depended_on_by, starting_nodes=var_changes
        )
        return nodes

    def get_upstream_nodes(
        self,
        final_vars: List[str],
        runtime_inputs: Dict[str, Any] = None,
        runtime_overrides: Dict[str, Any] = None,
    ) -> Tuple[Set[node.Node], Set[node.Node]]:
        """Given our function graph, and a list of desired output variables, returns the subgraph
        required to compute them.

        :param final_vars: the list of node names we want.
        :param runtime_inputs: runtime inputs to the DAG -- if not provided we will assume we're running at compile-time. Everything
        would then be required (even though it might be marked as optional), as we want to crawl
        the whole DAG. If we're at runtime, we want to just use the nodes that are provided,
        and not count the optional ones that are not.
        :param runtime_overrides: runtime overrides to the DAG -- if not provided we will assume
        we're running at compile-time.
        :return: a tuple of sets: - set of all nodes. - subset of nodes that human input is required for.
        """

        def next_nodes_function(n: node.Node) -> List[node.Node]:
            deps = []
            if runtime_overrides is not None and n.name in runtime_overrides:
                return deps
            if runtime_inputs is None:
                return n.dependencies
            for dep in n.dependencies:
                # If inputs is None, we want to assume its required, as it is a compile-time
                # dependency
                if (
                    dep.user_defined
                    and dep.name not in runtime_inputs
                    and dep.name not in self.config
                ):
                    _, dependency_type = n.input_types[dep.name]
                    if dependency_type == node.DependencyType.OPTIONAL:
                        continue
                deps.append(dep)
            return deps

        return self.directional_dfs_traverse(
            next_nodes_function, starting_nodes=final_vars, runtime_inputs=runtime_inputs
        )

    def nodes_between(self, start: str, end: str) -> Set[node.Node]:
        """Given our function graph, and a list of desired output variables, returns the subgraph
        required to compute them. Note that this returns an empty set if no path exists.


        :param start: the start of the subDAG we're looking for (inputs to it)
        :param end: the end of the subDAG we're looking for
        :return: The set of all nodes between start and end inclusive
        """

        end_node = self.nodes[end]
        start_node, between = graph_functions.nodes_between(
            self.nodes[end], lambda node_: node_.name == start
        )
        path_exists = start_node is not None
        if not path_exists:
            return set()
        return set(([start_node] if start_node is not None else []) + between + [end_node])

    def directional_dfs_traverse(
        self,
        next_nodes_fn: Callable[[node.Node], Collection[node.Node]],
        starting_nodes: List[str],
        runtime_inputs: Dict[str, Any] = None,
    ):
        """Traverses the DAG directionally using a DFS.

        :param next_nodes_fn: Function to give the next set of nodes
        :param starting_nodes: Which nodes to start at.
        :param runtime_inputs: runtime inputs to the DAG. This is here to allow for inputs to be also outputs.
        :return: a tuple of sets:
            - set of all nodes.
            - subset of nodes that human input is required for.
        """
        if runtime_inputs is None:
            runtime_inputs = {}
        nodes = set()
        user_nodes = set()

        def dfs_traverse(node: node.Node):
            nodes.add(node)
            for n in next_nodes_fn(node):
                if n not in nodes:
                    dfs_traverse(n)
            if node.user_defined:
                user_nodes.add(node)

        missing_vars = []
        for var in starting_nodes:
            if var not in self.nodes and var not in self.config:
                # checking for runtime_inputs because it's not in the graph isn't really a graph concern. So perhaps we
                # should move this outside of the graph in the future. This will do fine for now.
                if var not in runtime_inputs:
                    # if it's not in the runtime inputs, it's a properly missing variable
                    missing_vars.append(var)
                continue  # collect all missing final variables
            dfs_traverse(self.nodes[var])
        if missing_vars:
            missing_vars_str = ",\n".join(missing_vars)
            raise ValueError(f"Unknown nodes [{missing_vars_str}] requested. Check for typos?")
        return nodes, user_nodes

    def execute(
        self,
        nodes: Collection[node.Node] = None,
        computed: Dict[str, Any] = None,
        overrides: Dict[str, Any] = None,
        inputs: Dict[str, Any] = None,
        run_id: str = None,
    ) -> Dict[str, Any]:
        """Executes the DAG, given potential inputs/previously computed components.

        :param nodes: Nodes to compute
        :param computed: Nodes that have already been computed
        :param overrides: Overrides for nodes in the DAG
        :param inputs: Inputs to the DAG -- have to be disjoint from config.
        :return: The result of executing the DAG (a dict of node name to node result)
        """
        if nodes is None:
            nodes = self.get_nodes()
        if inputs is None:
            inputs = {}
        if run_id is None:
            run_id = str(uuid.uuid4())
        inputs = graph_functions.combine_config_and_inputs(self.config, inputs)
        return graph_functions.execute_subdag(
            nodes=nodes,
            inputs=inputs,
            adapter=self.adapter,
            computed=computed,
            overrides=overrides,
            run_id=run_id,
        )



from pathlib import Path

from hamilton import driver
from hamilton.cli import logic

from tests.cli.resources import module_v1, module_v2


def test_git_directory_exists():
    git_base_dir = logic.get_git_base_directory()

    assert Path(git_base_dir).exists()


def test_map_nodes_to_origins():
    expected_mapping = {
        "customers_path": "customers_df",
        "customers_df": "customers_df",
        "orders_path": "orders_df",
        "orders_df": "orders_df",
        "customers_orders_df": "customers_orders_df",
        "amount": "customers_orders_df",
        "age": "customers_orders_df",
        "country": "customers_orders_df",
        "orders_per_customer": "orders_per_customer",
        "average_order_by_customer": "average_order_by_customer",
        "customer_summary_table": "customer_summary_table",
    }

    dr = driver.Builder().with_modules(module_v1).build()
    node_to_origin = logic.map_nodes_to_functions(dr)

    assert node_to_origin == expected_mapping


def test_diff_versions():
    reference_versions = {
        "average_order_by_customer": "b58a6",
        "customer_summary_table": "6bf52",
        "customers_df": "480be",
        "customers_orders_df": "883f0",
        "orders_df": "58e65",
        "orders_per_customer": "6af6d",
    }
    current_versions = {
        "average_order_by_customer": "5296f",
        "customer_summary_table": "6bf52",
        "customers_df": "480be",
        "customers_orders_df": "883f0",
        "orders_df": "58e65",
        "orders_per_distributor": "6d64l",
    }

    diff = logic.diff_versions(
        current_map=current_versions,
        reference_map=reference_versions,
    )

    assert diff["reference_only"] == ["orders_per_customer"]
    assert diff["current_only"] == ["orders_per_distributor"]
    assert diff["edit"] == ["average_order_by_customer"]


def test_diff_node_versions():
    current_dr = driver.Builder().with_modules(module_v2).build()
    reference_dr = driver.Builder().with_modules(module_v1).build()

    current_nodes = logic.hash_hamilton_nodes(current_dr)
    reference_nodes = logic.hash_hamilton_nodes(reference_dr)

    diff = logic.diff_versions(
        current_map=current_nodes,
        reference_map=reference_nodes,
    )

    assert diff["reference_only"] == ["orders_per_customer"]
    assert diff["current_only"] == ["orders_per_distributor"]
    assert diff["edit"] == ["average_order_by_customer", "customer_summary_table"]



import logging

import hamilton.async_driver

logger = logging.getLogger(__name__)

logger.warning(
    "This module is deprecated and will be removed in Hamilton 2.0 "
    "Please use `hamilton.async_driver` instead. "
)

AsyncDriver = hamilton.async_driver.AsyncDriver



import inspect
import logging
import typing
from typing import Any, Callable, Collection, Dict, List, Optional, Tuple, Type

from hamilton import node
from hamilton.function_modifiers.base import (
    InvalidDecoratorException,
    NodeInjector,
    SingleNodeNodeTransformer,
)
from hamilton.function_modifiers.dependencies import (
    LiteralDependency,
    ParametrizedDependency,
    UpstreamDependency,
)
from hamilton.io.data_adapters import AdapterCommon, DataLoader, DataSaver
from hamilton.node import DependencyType
from hamilton.registry import LOADER_REGISTRY, SAVER_REGISTRY

logger = logging.getLogger(__name__)


class AdapterFactory:
    """Factory for data loaders. This handles the fact that we pass in source(...) and value(...)
    parameters to the data loaders."""

    def __init__(self, adapter_cls: Type[AdapterCommon], **kwargs: ParametrizedDependency):
        """Initializes an adapter factory. This takes in parameterized dependencies
        and stores them for later resolution.

        Note that this is not strictly necessary -- we could easily put this in the
        decorator, but I wanted to separate out/consolidate the logic between data savers and data
        loaders.

        :param adapter_cls: Class of the loader to create.
        :param kwargs: Keyword arguments to pass to the loader, as parameterized dependencies.
        """
        self.adapter_cls = adapter_cls
        self.kwargs = kwargs
        self.validate()

    def validate(self):
        """Validates that the loader class has the required arguments, and that
        the arguments passed in are valid.

        :raises InvalidDecoratorException: If the arguments are invalid.
        """
        required_args = self.adapter_cls.get_required_arguments()
        optional_args = self.adapter_cls.get_optional_arguments()
        missing_params = set(required_args.keys()) - set(self.kwargs.keys())
        extra_params = (
            set(self.kwargs.keys()) - set(required_args.keys()) - set(optional_args.keys())
        )
        if len(missing_params) > 0:
            raise InvalidDecoratorException(
                f"Missing required parameters for adapter : {self.adapter_cls}: {missing_params}. "
                f"Required parameters/types are: {required_args}. Optional parameters/types are: "
                f"{optional_args}. "
            )
        if len(extra_params) > 0:
            available_args = {**required_args, **optional_args}.keys()
            raise InvalidDecoratorException(
                f"Extra parameters for loader: {self.adapter_cls} {extra_params}. Choices for parameters are: "
                f"{available_args}."
            )

    def create_loader(self, **resolved_kwargs: Any) -> DataLoader:
        if not self.adapter_cls.can_load():
            raise InvalidDecoratorException(f"Adapter {self.adapter_cls} cannot load data.")
        return self.adapter_cls(**resolved_kwargs)

    def create_saver(self, **resolved_kwargs: Any) -> DataSaver:
        if not self.adapter_cls.can_save():
            raise InvalidDecoratorException(f"Adapter {self.adapter_cls} cannot save data.")
        return self.adapter_cls(**resolved_kwargs)


def resolve_kwargs(kwargs: Dict[str, Any]) -> Tuple[Dict[str, str], Dict[str, Any]]:
    """Resolves kwargs to a list of dependencies, and a dictionary of name
    to resolved literal values.

    :return: A tuple of the dependencies, and the resolved literal kwargs.
    """
    dependencies = {}
    resolved_kwargs = {}
    for name, dependency in kwargs.items():
        if isinstance(dependency, UpstreamDependency):
            dependencies[name] = dependency.source
        elif isinstance(dependency, LiteralDependency):
            resolved_kwargs[name] = dependency.value
        else:
            resolved_kwargs[name] = dependency
    return dependencies, resolved_kwargs


def resolve_adapter_class(
    type_: Type[Type], loader_classes: List[Type[AdapterCommon]]
) -> Optional[Type[AdapterCommon]]:
    """Resolves the loader class for a function. This will return the most recently
    registered loader class that applies to the injection type, hence the reversed order.

    :param fn: Function to inject the loaded data into.
    :return: The loader class to use.
    """
    applicable_adapters: List[Type[AdapterCommon]] = []
    loaders_with_any = []
    for loader_cls in reversed(loader_classes):
        # We do this here, rather than in applies_to, as its a bit of a special case
        # Any should always get last priority, as its very non-specific and should be able to
        # This is partially for backwards compatibility as we haven't always supported these -- including it now
        # would potentially use the wrong loader for production cases
        if Any in loader_cls.applicable_types():
            loaders_with_any.append(loader_cls)
        if loader_cls.applies_to(type_):
            applicable_adapters.append(loader_cls)
    if len(applicable_adapters) > 0:
        if len(applicable_adapters) > 1:
            logger.warning(
                f"More than one applicable adapter detected for {type_}. "
                f"Using the last one registered {applicable_adapters[0]}."
            )
        return applicable_adapters[0]
    if loaders_with_any:
        return loaders_with_any[0]
    return None


class LoadFromDecorator(NodeInjector):
    def __init__(
        self,
        loader_classes: typing.Sequence[Type[DataLoader]],
        inject_=None,
        **kwargs: ParametrizedDependency,
    ):
        """Instantiates a load_from decorator. This decorator will load from a data source,
        and

        :param inject: The name of the parameter to inject the data into.
        :param loader_cls: The data loader class to use.
        :param kwargs: The arguments to pass to the data loader.
        """
        self.loader_classes = loader_classes
        self.kwargs = kwargs
        self.inject = inject_

    def _select_param_to_inject(self, params: List[str], fn: Callable) -> str:
        """Chooses a parameter to inject, given the parameters available. If self.inject is None
        (meaning we inject the only parameter), then that's the one. If it is not None, then
        we need to ensure it is one of the available parameters, in which case we choose it.
        """
        if self.inject is None:
            if len(params) == 1:
                return params[0]
            raise InvalidDecoratorException(
                f"If the nodes produced by {fn.__qualname__} require multiple inputs, "
                f"you must pass `inject_` to the load_from decorator for "
                f"function: {fn.__qualname__}"
            )
        if self.inject not in params:
            raise InvalidDecoratorException(
                f"Parameter {self.inject} not required by nodes produced by {fn.__qualname__}"
            )
        return self.inject

    def get_loader_nodes(
        self, inject_parameter: str, load_type: Type[Type], namespace: str = None
    ) -> List[node.Node]:
        loader_cls = resolve_adapter_class(
            load_type,
            self.loader_classes,
        )
        if loader_cls is None:
            raise InvalidDecoratorException(
                f"Could not resolve loader for type: {load_type} given possibilities: {self.loader_classes} "
            )
        loader_factory = AdapterFactory(loader_cls, **self.kwargs)
        # dependencies is a map from param name -> source name
        # we use this to pass the right arguments to the loader.
        dependencies, resolved_kwargs = resolve_kwargs(self.kwargs)
        # we need to invert the dependencies so that we can pass
        # the right argument to the loader
        dependencies_inverted = {v: k for k, v in dependencies.items()}

        def load_data(
            __loader_factory: AdapterFactory = loader_factory,
            __load_type: Type[Type] = load_type,
            __resolved_kwargs=resolved_kwargs,
            __dependencies=dependencies_inverted,
            __optional_params=loader_cls.get_optional_arguments(),
            **input_kwargs: Any,
        ) -> Tuple[load_type, Dict[str, Any]]:
            input_args_with_fixed_dependencies = {
                __dependencies.get(key, key): value for key, value in input_kwargs.items()
            }
            kwargs = {**__resolved_kwargs, **input_args_with_fixed_dependencies}
            data_loader = __loader_factory.create_loader(**kwargs)
            return data_loader.load_data(load_type)

        def get_input_type_key(key: str) -> str:
            return key if key not in dependencies else dependencies[key]

        input_types = {
            get_input_type_key(key): (type_, DependencyType.REQUIRED)
            for key, type_ in loader_cls.get_required_arguments().items()
        }
        input_types.update(
            {
                dependencies[key]: (type_, DependencyType.OPTIONAL)
                for key, type_ in loader_cls.get_optional_arguments().items()
                if key in dependencies
            }
        )
        # Take out all the resolved kwargs, as they are not dependencies, and will be filled out
        # later
        input_types = {
            key: value for key, value in input_types.items() if key not in resolved_kwargs
        }

        # the loader node is the node that loads the data from the data source.
        loader_node = node.Node(
            name=f"{inject_parameter}",
            callabl=load_data,
            typ=Tuple[Dict[str, Any], load_type],
            input_types=input_types,
            tags={
                "hamilton.data_loader": True,
                "hamilton.data_loader.has_metadata": True,
                "hamilton.data_loader.source": f"{loader_cls.name()}",
                "hamilton.data_loader.classname": f"{loader_cls.__qualname__}",
                "hamilton.data_loader.node": inject_parameter,
            },
            namespace=(
                (namespace, "load_data") if namespace else ("load_data",)
            ),  # We want no namespace in this case
        )

        # the filter node is the node that takes the data from the data source, filters out
        # metadata, and feeds it in as a parameter. Note that this must have the same name as the
        # inject parameter

        def filter_function(_inject_parameter=inject_parameter, **kwargs):
            return kwargs[loader_node.name][0]  # filter it out

        filter_node = node.Node(
            name=f"{inject_parameter}",
            callabl=filter_function,
            typ=load_type,
            input_types={loader_node.name: loader_node.type},
            tags={
                "hamilton.data_loader": True,
                "hamilton.data_loader.has_metadata": False,
                "hamilton.data_loader.source": f"{loader_cls.name()}",
                "hamilton.data_loader.classname": f"{loader_cls.__qualname__}",
                "hamilton.data_loader.node": inject_parameter,
            },
            # This is a little sloppy -- we need to figure out the best way to handle namespacing
            # In reality we will likely be changing the API -- using the logging construct so we don't have
            # to have this weird DAG shape. For now, this solves the problem, and this is an internal component of the API
            # so we're good to go
            namespace=((namespace, "select_data") if namespace else ()),  # We want no namespace
        )
        return [loader_node, filter_node]

    def inject_nodes(
        self, params: Dict[str, Type[Type]], config: Dict[str, Any], fn: Callable
    ) -> Tuple[Collection[node.Node], Dict[str, str]]:
        pass
        """Generates two nodes:
        1. A node that loads the data from the data source, and returns that + metadata
        2. A node that takes the data from the data source, injects it into, and runs, the function.

        :param fn: The function to decorate.
        :param config: The configuration to use.
        :return: The resolved nodes
        """
        inject_parameter = self._select_param_to_inject(list(params.keys()), fn)
        load_type = params[inject_parameter]
        try:
            loader_node, filter_node = self.get_loader_nodes(
                inject_parameter, load_type, fn.__name__
            )
        except InvalidDecoratorException as e:
            raise InvalidDecoratorException(f"Cannot find : {fn.__qualname__}: {e}") from e

        return [loader_node, filter_node], {inject_parameter: filter_node.name}

    def _get_inject_parameter_from_function(self, fn: Callable) -> Tuple[str, Type[Type]]:
        """Gets the name of the parameter to inject the data into.

        :param fn: The function to decorate.
        :return: The name of the parameter to inject the data into.
        """
        sig = inspect.signature(fn)
        if self.inject is None:
            if len(sig.parameters) == 0:
                raise InvalidDecoratorException(
                    f"The function: {fn.__qualname__} has no parameters. "
                    f"The data loader functionality injects the loaded data "
                    f"into the function, so you must have at least one parameter."
                )
            if len(sig.parameters) != 1:
                raise InvalidDecoratorException(
                    f"If you have multiple parameters in the signature, "
                    f"you must pass `inject_` to the load_from decorator for "
                    f"function: {fn.__qualname__}"
                )
            inject = list(sig.parameters.keys())[0]

        else:
            if self.inject not in sig.parameters:
                raise InvalidDecoratorException(
                    f"Invalid inject parameter: {self.inject} for fn: {fn.__qualname__}"
                )
            inject = self.inject
        return inject, typing.get_type_hints(fn)[inject]

    def validate(self, fn: Callable):
        """Note that we actually do something a little clever here. While the decorator operates
        on the nodes in the subdag, we actually validate it against the function. This ensures that
        we're not doing a lot of weird dynamic things that inject parameters that aren't in the
        function.

        :param fn: Function that produces the data loader.
        """

        inject_parameter, type_ = self._get_inject_parameter_from_function(fn)
        cls = resolve_adapter_class(type_, self.loader_classes)
        if cls is None:
            raise InvalidDecoratorException(
                f"No loader class found for type: {type_} specified by "
                f"parameter: {inject_parameter} in function: {fn.__qualname__}"
            )
        loader_factory = AdapterFactory(cls, **self.kwargs)
        loader_factory.validate()


class load_from__meta__(type):
    """Metaclass for the load_from decorator. This is specifically to allow class access method.
    Note that there *is* another way to do this -- we couold add attributes dynamically on the
    class in registry, or make it a function that just proxies to the decorator. We can always
    change this up, but this felt like a pretty clean way of doing it, where we can decouple the
    registry from the decorator class.
    """

    def __getattr__(cls, item: str):
        if item in LOADER_REGISTRY:
            return load_from.decorator_factory(LOADER_REGISTRY[item])
        try:
            return super().__getattribute__(item)
        except AttributeError as e:
            raise AttributeError(
                f"No loader named: {item} available for {cls.__name__}. "
                f"Available loaders are: {LOADER_REGISTRY.keys()}. "
                f"If you've gotten to this point, you either (1) spelled the "
                f"loader name wrong, (2) are trying to use a loader that does"
                f"not exist (yet). For a list of available loaders, see: "
                f"https://hamilton.readthedocs.io/reference/io/available-data-adapters/#data"
                f"-loaders "
            ) from e


class load_from(metaclass=load_from__meta__):
    """Decorator to inject externally loaded data into a function. Ideally, anything that is not
    a pure transform should either call this, or accept inputs from an external location.

    This decorator functions by "injecting" a parameter into the function. For example,
    the following code will load the json file, and inject it into the function as the parameter
    `input_data`. Note that the path for the JSON file comes from another node called
    raw_data_path (which could also be passed in as an external input).

    .. code-block:: python

        @load_from.json(path=source("raw_data_path"))
        def raw_data(input_data: dict) -> dict:
            return input_data

    The decorator can also be used with `value` to inject a constant value into the loader.
    In the following case, we use the literal value "some/path.json" as the path to the JSON file.

    .. code-block:: python

        @load_from.json(path=value("some/path.json"))
        def raw_data(input_data: dict) -> dict:
            return input_data

    Note that, if neither `source` nor `value` is specified, the value will be passed in as a
    literal value.

    .. code-block:: python

        @load_from.json(path="some/path.json")
        def raw_data(input_data: dict) -> dict:
            return input_data

    You can also utilize the `inject_` parameter in the loader if you want to inject the data
    into a specific param. For example, the following code will load the json file, and inject it
    into the function as the parameter `data`.

    .. code-block:: python

        @load_from.json(path=source("raw_data_path"), inject_="data")
        def raw_data(data: dict, valid_keys: List[str]) -> dict:
            return [item for item in data if item in valid_keys]

    You can also utilize multiple data loaders with separate `inject_` parameters to
    load from multiple files.
    data loaders to a single function:

    .. code-block:: python

        @load_from.json(path=source("raw_data_path"), inject_="data")
        @load_from.json(path=source("raw_data_path2"), inject_="data2")
        def raw_data(data: dict, data2: dict) -> dict:
            return [item for item in data if item in data2]


    This is a highly pluggable functionality -- here's the basics of how it works:

    1. Every "key" (json above, but others include csv, literal, file, pickle, etc...) corresponds
    to a set of loader classes. For example, the json key corresponds to the JSONLoader class in
    default_data_loaders. They implement the classmethod `name`. Once they are registered with the
    central registry they pick

    2. Every data loader class (which are all dataclasses) implements the `load_targets` method,
    which returns a list of types it can load to. For example, the JSONLoader class can load data
    of type `dict`. Note that the set of potential loading candidate classes are evaluated in
    reverse order, so the most recently registered loader class is the one that is used. That
    way, you can register custom ones.

    3. The loader class is instantiated with the kwargs passed to the decorator. For example, the
    JSONLoader class takes a `path` kwarg, which is the path to the JSON file.

    4. The decorator then creates a node that loads the data, and modifies the node that runs the
    function to accept that. It also returns metadata (customizable at the loader-class-level) to
    enable debugging after the fact. This is unstructured, but can be used down the line to describe
    any metadata to help debug.

    The "core" hamilton library contains a few basic data loaders that can be implemented within
    the confines of python's standard library. pandas_extensions contains a few more that require
    pandas to be installed.

    Note that these can have `default` arguments, specified by defaults in the dataclass fields.
    For the full set of "keys" and "types" (e.g. load_from.json, etc...), look for all classes
    that inherit from `DataLoader` in the hamilton library. We plan to improve documentation shortly
    to make this discoverable.
    """

    def __call__(self, *args, **kwargs):
        return LoadFromDecorator(*args, **kwargs)

    @classmethod
    def decorator_factory(
        cls, loaders: typing.Sequence[Type[DataLoader]]
    ) -> Callable[..., LoadFromDecorator]:
        """Effectively a partial function for the load_from decorator. Broken into its own (
        rather than using functools.partial) as it is a little clearer to parse.

        :param loaders: Options of data loader classes to use
        :return: The data loader decorator.
        """

        def create_decorator(
            __loaders=tuple(loaders), inject_=None, **kwargs: ParametrizedDependency
        ):
            return LoadFromDecorator(__loaders, inject_=inject_, **kwargs)

        return create_decorator


class save_to__meta__(type):
    """See note on load_from__meta__ for details on how this works."""

    def __getattr__(cls, item: str):
        if item in SAVER_REGISTRY:
            return save_to.decorator_factory(SAVER_REGISTRY[item])
        try:
            return super().__getattribute__(item)
        except AttributeError as e:
            raise AttributeError(
                f"No saver named: {item} available for {cls.__name__}. "
                f"Available data savers are: {list(SAVER_REGISTRY.keys())}. "
                "If you've gotten to this point, you either (1) spelled the "
                "loader name wrong, (2) are trying to use a saver that does"
                "not exist (yet). For a list of available savers, see "
                "https://hamilton.dagworks.io/en/latest/reference/io/available-data-adapters/"
            ) from e


class SaveToDecorator(SingleNodeNodeTransformer):
    def __init__(
        self,
        saver_classes_: typing.Sequence[Type[DataSaver]],
        output_name_: str = None,
        target_: str = None,
        **kwargs: ParametrizedDependency,
    ):
        super(SaveToDecorator, self).__init__()
        self.artifact_name = output_name_
        self.saver_classes = saver_classes_
        self.kwargs = kwargs
        self.target = target_

    def create_saver_node(
        self, node_: node.Node, config: Dict[str, Any], fn: Callable
    ) -> node.Node:
        artifact_name = self.artifact_name
        artifact_namespace = ()
        node_to_save = node_.name if self.target is None else self.target

        if artifact_name is None:
            artifact_name = node_to_save
            artifact_namespace = ("save",)

        type_ = node_.type
        saver_cls = resolve_adapter_class(
            type_,
            self.saver_classes,
        )
        if saver_cls is None:
            raise InvalidDecoratorException(
                f"No saver class found for type: {type_} specified by "
                f"output type: {type_} in node: {node_to_save} generated by "
                f"function: {fn.__qualname__}."
            )

        adapter_factory = AdapterFactory(saver_cls, **self.kwargs)
        dependencies, resolved_kwargs = resolve_kwargs(self.kwargs)
        dependencies_inverted = {v: k for k, v in dependencies.items()}

        def save_data(
            __adapter_factory=adapter_factory,
            __dependencies=dependencies_inverted,
            __resolved_kwargs=resolved_kwargs,
            __data_node_name=node_to_save,
            **input_kwargs,
        ) -> Dict[str, Any]:
            input_args_with_fixed_dependencies = {
                __dependencies.get(key, key): value for key, value in input_kwargs.items()
            }
            kwargs = {**__resolved_kwargs, **input_args_with_fixed_dependencies}
            data_to_save = kwargs[__data_node_name]
            kwargs = {k: v for k, v in kwargs.items() if k != __data_node_name}
            data_saver = __adapter_factory.create_saver(**kwargs)
            return data_saver.save_data(data_to_save)

        def get_input_type_key(key: str) -> str:
            return key if key not in dependencies else dependencies[key]

        input_types = {
            get_input_type_key(key): (type_, DependencyType.REQUIRED)
            for key, type_ in saver_cls.get_required_arguments().items()
        }
        input_types.update(
            {
                dependencies[key]: (type_, DependencyType.OPTIONAL)
                for key, type_ in saver_cls.get_optional_arguments().items()
                if key in dependencies
            }
        )
        # Take out all the resolved kwargs, as they are not dependencies, and will be filled out
        # later
        input_types = {
            key: value for key, value in input_types.items() if key not in resolved_kwargs
        }
        input_types[node_to_save] = (node_.type, DependencyType.REQUIRED)
        save_node = node.Node(
            name=artifact_name,
            callabl=save_data,
            typ=Dict[str, Any],
            input_types=input_types,
            namespace=artifact_namespace,
            tags={
                "hamilton.data_saver": True,
                "hamilton.data_saver.sink": f"{saver_cls.name()}",
                "hamilton.data_saver.classname": f"{saver_cls.__qualname__}",
            },
        )
        return save_node

    def transform_node(
        self, node_: node.Node, config: Dict[str, Any], fn: Callable
    ) -> Collection[node.Node]:
        """Transforms the node to a data saver.

        :param node_: Node to transform
        :param config: Config to use
        :param fn: Original function that generated the node
        :return: The transformed node
        """

        return [
            self.create_saver_node(node_, config, fn),
            node_,
        ]

    def validate(self, fn: Callable):
        pass


class save_to(metaclass=save_to__meta__):
    """Decorator that outputs data to some external source. You can think
    about this as the inverse of load_from.

    This decorates a function, takes the final node produced by that function
    and then appends an additional node that saves the output of that function.

    As the load_from decorator does, this decorator can be referred to in a
    dynamic way. For instance, @save_to.json will save the output of the function
    to a json file. Note that this means that the output of the function must
    be a dictionary (or subclass thereof), otherwise the decorator will fail.

    Looking at the json example:

    .. code-block:: python

        @save_to.json(path=source("raw_data_path"), output_name_="data_save_output")
        def final_output(data: dict, valid_keys: List[str]) -> dict:
            return [item for item in data if item in valid_keys]

    This adds a final node to the DAG with the name "data_save_output" that
    accepts the output of the function "final_output" and saves it to a json.
    In this case, the JSONSaver accepts a `path` parameter, which is provided
    by the upstream node (or input) named "raw_data_path". The `output_name_`
    parameter then says how to refer to the output of this node in the DAG.

    If you called this with the driver:

    .. code-block:: python

        dr = driver.Driver(my_module)
        output = dr.execute(['final_output'], {'raw_data_path': '/path/my_data.json'})

    You would *just* get the final result, and nothing would be saved.

    If you called this with the driver:

    .. code-block:: python

        dr = driver.Driver(my_module)
        output = dr.execute(['data_save_output'], {'raw_data_path': '/path/my_data.json'})

    You would get a dictionary of metadata (about the saving output), and the final result would
    be saved to a path.

    Note that you can also hardcode the path, rather than using a dependency:

    .. code-block:: python

        @save_to.json(path=value('/path/my_data.json'), output_name_="data_save_output")
        def final_output(data: dict, valid_keys: List[str]) -> dict:
            return [item for item in data if item in valid_keys]

    Note that, like the loader function, you can use literal values as kwargs and they'll get interpreted as values.
    If you needs savers, you should also look into `.materialize` on the driver -- it's a clean way to do this in a more
    ad-hoc/decoupled manner.

    If you want to layer savers, you'll have to use the target\\_ parameter, which tells the saver which node to use.

    .. code-block:: python

        @save_to.json(path=source("raw_data_path"), output_name_="data_save_output", target_="data")
        @save_to.json(path=source("raw_data_path2"), output_name_="data_save_output2", target_="data")
        def final_output(data: dict, valid_keys: List[str]) -> dict:
            return [item for item in data if item in valid_keys]

    """

    def __call__(self, *args, **kwargs):
        return SaveToDecorator(*args, **kwargs)

    @classmethod
    def decorator_factory(
        cls, savers: typing.Sequence[Type[DataSaver]]
    ) -> Callable[..., SaveToDecorator]:
        """Effectively a partial function for the load_from decorator. Broken into its own (
        rather than using functools.partial) as it is a little clearer to parse.

        :param savers: Candidate data savers
        :param loaders: Options of data loader classes to use
        :return: The data loader decorator.
        """

        def create_decorator(__savers=tuple(savers), **kwargs: ParametrizedDependency):
            return SaveToDecorator(__savers, **kwargs)

        return create_decorator



import traceback
from typing import Any, Dict, Optional

from slack_sdk import WebClient

from hamilton.lifecycle import NodeExecutionHook


class SlackNotifier(NodeExecutionHook):
    """This is a adapter that sends a message to a slack channel when a node is executed.

    Note: you need to have slack_sdk installed for this to work.
    If you don't have it installed, you can install it with `pip install slack_sdk`
    (or `pip install sf-hamilton[slack]` -- use quotes if you're using zsh).

    .. code-block:: python

        from hamilton.plugins import h_slack

        dr = (
            driver.Builder()
            .with_config({})
            .with_modules(some_modules)
            .with_adapters(h_slack.SlackNotifier(api_key="YOUR_API_KEY", channel="YOUR_CHANNEL"))
            .build()
        )
        # and then when you call .execute() or .materialize() you'll get a message in your slack channel!

    """

    def __init__(self, api_key: str, channel: str, **kwargs):
        """Constructor.

        :param api_key: API key to use for sending messages.
        :param channel: Channel to send messages to.
        """
        self.slack_client = WebClient(api_key)
        self.channel = channel
        self.kwargs = kwargs

    def _send_message(self, message: str):
        """Sends a message to the slack channel."""
        if self.slack_client is not None:
            self.slack_client.chat_postMessage(channel=self.channel, text=message)

    def run_before_node_execution(
        self,
        node_name: str,
        node_tags: Dict[str, Any],
        node_kwargs: Dict[str, Any],
        node_return_type: type,
        **future_kwargs: Any,
    ):
        """Placeholder required to subclass `NodeExecutionMethod`"""
        pass

    def run_after_node_execution(
        self,
        node_name: str,
        node_tags: Dict[str, Any],
        node_kwargs: Dict[str, Any],
        node_return_type: type,
        result: Any,
        error: Optional[Exception],
        success: bool,
        task_id: Optional[str],
        run_id: str,
        **future_kwargs: Any,
    ):
        """Sends a message to the slack channel after a node is executed."""
        if error is not None:
            message = f"Error in Executing Node: {node_name}. Task ID: {task_id}. Run ID: {run_id}. Error: {error}.\n Stack Trace: {''.join(traceback.format_tb(error.__traceback__))}"
            self._send_message(message)



import inspect
import sys
import typing
from abc import ABC
from typing import Any, Generator, Optional, Tuple, Type, TypeVar, Union

import typing_inspect

if sys.version_info >= (3, 9):
    from typing import Literal
else:
    Literal = None
from hamilton.registry import COLUMN_TYPE, DF_TYPE_AND_COLUMN_TYPES

BASE_ARGS_FOR_GENERICS = (typing.T,)


def _safe_subclass(candidate_type: Type, base_type: Type) -> bool:
    """Safely checks subclass, returning False if python's subclass does not work.
    This is *not* a true subclass check, and will not tell you whether hamilton
    considers the types to be equivalent. Rather, it is used to short-circuit further
    computation safely and avoid errors.

    Note that we may end up with types that *should* be considered equivalent, but
    are not. In that case we will deal with them -- its a better user experience and easier
    to report than an error.

    :param base_type: Base type to check against
    :param candidate_type: Candidate type to check as a potential subclass
    :return: Whether python considers them subclasses and will not break if subclass is called.
    """
    if len(_get_args(candidate_type)) > 0 or len(_get_args(base_type)) > 0:
        return False
    if inspect.isclass(candidate_type) and inspect.isclass(base_type):
        return issubclass(candidate_type, base_type)
    return False


def custom_subclass_check(requested_type: Type, param_type: Type):
    """This is a custom check around generics & classes. It probably misses a few edge cases.

    We will likely need to revisit this in the future (perhaps integrate with graphadapter?)

    :param requested_type: Candidate subclass.
    :param param_type: Type of parameter to check against.
    :return: Whether or not requested_type is a valid subclass of param_type.
    """
    # handles case when someone is using primitives and generics
    requested_origin_type = requested_type
    param_type, _ = get_type_information(param_type)
    param_origin_type = param_type
    has_generic = False
    if param_type == Any:
        # any type is a valid subclass of Any.
        return True
    if _safe_subclass(requested_type, param_type):
        return True
    if typing_inspect.is_union_type(param_type):
        for arg in _get_args(param_type):
            if custom_subclass_check(requested_type, arg):
                return True
    if typing_inspect.is_generic_type(requested_type) or typing_inspect.is_tuple_type(
        requested_type
    ):
        requested_origin_type = typing_inspect.get_origin(requested_type)
        has_generic = True
    if typing_inspect.is_generic_type(param_type) or typing_inspect.is_tuple_type(param_type):
        param_origin_type = typing_inspect.get_origin(param_type)
        has_generic = True
    # TODO -- consider moving into a graph adapter or elsewhere -- this is perhaps a little too
    #  low-level
    if has_generic and requested_origin_type in (Parallelizable,):
        (requested_type_arg,) = _get_args(requested_type)
        return custom_subclass_check(requested_type_arg, param_type)
    if has_generic and param_origin_type == Collect:
        (param_type_arg,) = _get_args(param_type)
        return custom_subclass_check(requested_type, param_type_arg)
    if requested_origin_type == param_origin_type or _safe_subclass(
        requested_origin_type, param_origin_type
    ):
        if has_generic:  # check the args match or they do not have them defined.
            requested_args = _get_args(requested_type)
            param_args = _get_args(param_type)
            if (
                requested_args
                and param_args
                and requested_args != BASE_ARGS_FOR_GENERICS
                and param_args != BASE_ARGS_FOR_GENERICS
            ):
                return requested_args == param_args
        return True
    return False


def get_type_as_string(type_: Type) -> Optional[str]:
    """Get a string representation of a type.

    The logic supports the evolution of the type system between 3.8 and 3.10.
    :param type_: Any Type object. Typically the node type found at Node.type.
    :return: string representation of the type. An empty string if everything fails.
    """

    if getattr(type_, "__name__", None):
        type_string = type_.__name__
    elif typing_inspect.get_origin(type_):
        base_type = typing_inspect.get_origin(type_)
        type_string = get_type_as_string(base_type)
    elif getattr(type_, "__repr__", None):
        type_string = type_.__repr__()
    else:
        type_string = None

    return type_string


def types_match(param_type: Type[Type], required_node_type: Any) -> bool:
    """Checks that we have "types" that "match".

    Matching can be loose here -- and depends on the adapter being used as to what is
    allowed. Otherwise it does a basic equality check.

    :param adapter: the graph adapter to delegate to for one check.
    :param param_type: the parameter type we're checking.
    :param required_node_type: the expected parameter type to validate against.
    :return: True if types are "matching", False otherwise.
    """
    if required_node_type == typing.Any:
        return True
    # type var  -- straight == should suffice. Assume people understand what they're doing with TypeVar.
    elif typing_inspect.is_typevar(required_node_type) or typing_inspect.is_typevar(param_type):
        return required_node_type == param_type
    elif required_node_type == param_type:
        return True
    elif custom_subclass_check(required_node_type, param_type):
        return True
    return False


_sys_version_info = sys.version_info
_version_tuple = (_sys_version_info.major, _sys_version_info.minor, _sys_version_info.micro)

"""
The following is purely for backwards compatibility
The behavior of annotated/get_args/get_origin has changed in recent versions
So we have to handle it accordingly
In 3.8/below we have to use the typing_extensions version

Also, note that it is currently called `column`, but
we will eventually want more options. E.G.

`dataset`
`scalar`
`tensor`

etc...

To do this, we'll likely extend from annotated, and add new types.
See `annotated` source code: https://github.com/python/cpython/blob/3.11/Lib/typing.py#L2122.

We can also potentially add validation in the types, and remove it from the validate.
"""

ANNOTATE_ALLOWED = False
if _version_tuple < (3, 9, 0):
    # Before 3.9 we use typing_extensions
    import typing_extensions

    column = typing_extensions.Annotated


else:
    ANNOTATE_ALLOWED = True
    from typing import Annotated, Type

    column = Annotated

if _version_tuple < (3, 9, 0):
    import typing_extensions

    _get_origin = typing_extensions.get_origin
    _get_args = typing_extensions.get_args
else:
    from typing import get_args as _get_args
    from typing import get_origin as _get_origin


def _is_annotated_type(type_: Type[Type]) -> bool:
    """Utility function to tell if a type is Annotated"""
    return _get_origin(type_) == column


# Placeholder exception for invalid hamilton types
class InvalidTypeException(Exception):
    pass


# Some valid series annotations
# We will likely have to expand
_valid_series_annotations = (
    int,
    float,
    str,
    bool,
)


def _is_valid_series_type(candidate_type: Type[Type]) -> bool:
    """Tells if something is a valid series type, using the registry we have.

    :param candidate_type: Type to check
    :return: Whether it is a series (column) type that we have registered
    """
    for key, types in DF_TYPE_AND_COLUMN_TYPES.items():
        if COLUMN_TYPE not in types:
            continue
        if issubclass(candidate_type, types[COLUMN_TYPE]):
            return True
    return False


def validate_type_annotation(annotation: Type[Type]):
    """Validates a type annotation for a hamilton function.
    If it is not an Annotated type, it will be fine.
    If it is the Annotated type, it will check that
    it only has one type annotation and that that is valid (currently int, float, str, bool).

    :param annotation: Annotation (e.g. Annotated[pd.Series, int])
    :raises InvalidTypeException: If the annotation is invalid
    """

    if not _is_annotated_type(annotation):
        # In this case we don't care too much -- hamilton accepts anything
        return True
    original, *annotations = _get_args(annotation)
    # TODO -- use extensions/series types to do this more effectively
    if not (_is_valid_series_type(original)):
        raise InvalidTypeException(
            f"Hamilton only accepts annotated types of series or equivalent. Got {original}"
        )
    if len(annotations) > 1 or len(annotations) == 0:
        raise InvalidTypeException(
            f"Hamilton only accepts one annotation per pd.Series. Got {annotations}"
        )
    subclasses_valid_annotation = False
    (annotation,) = annotations
    for valid_annotation in _valid_series_annotations:
        if custom_subclass_check(annotation, valid_annotation):
            subclasses_valid_annotation = True
    if not subclasses_valid_annotation:
        raise InvalidTypeException(
            f"Hamilton only accepts annotations on series that are subclasses of one of {_valid_series_annotations}. "
            f"Got {annotation}"
        )


def get_type_information(some_type: Any) -> Tuple[Type[Type], list]:
    """Gets the type information for a given type.

    If it is an annotated type, it will return the original type and the annotation.
    If it is not an annotated type, it will return the type and empty list.

    :param some_type: Type to get information for
    :return: Tuple of type and list of annotations (or empty list)
    """
    if _is_annotated_type(some_type):
        original, *annotations = _get_args(some_type)
        return original, annotations
    return some_type, []


# Type variables for annotations below
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


# TODO -- support sequential operation
# class Sequential(Generator[T, None, None], ABC):
#     pass


class Parallelizable(typing.Generator[U, None, None], ABC):
    pass


def is_parallelizable_type(type_: Type) -> bool:
    return issubclass(type_, Parallelizable)


class Collect(Generator[V, None, None], ABC):
    pass


def check_input_type(node_type: Type, input_value: Any) -> bool:
    """Checks an input value against the declare input type. This is a utility function to be
    used for checking types against values. Note we are looser here than in custom_subclass_check,
    as runtime-typing is less specific.

    :param node_type: Type of the node to check against.
    :param input_value: Value to check.
    :return: True if the input value is of the correct type, False otherwise.
    """
    if node_type == Any:
        return True
    # In the case of dict[str, Any] (or equivalent) in python 3.9 +
    # we need to double-check that its not generic, as the isinstance clause will break this
    elif (
        inspect.isclass(node_type)
        and not typing_inspect.is_generic_type(node_type)
        and isinstance(input_value, node_type)
    ):
        return True
    elif typing_inspect.is_typevar(node_type):  # skip runtime comparison for now.
        return True
    elif typing_inspect.is_generic_type(node_type) and typing_inspect.get_origin(node_type) == type(
        input_value
    ):
        return True
    elif typing_inspect.is_union_type(node_type):
        union_types = typing_inspect.get_args(node_type)
        return any([check_input_type(ut, input_value) for ut in union_types])
    elif node_type == type(input_value):
        return True
    return False


# TODO: merge the above with this in some way. Right now they're separate because they have different
# behaviors. We should determine how to reconcile these and how further type checking capabilities,
# e.g. handling Annotated, Pandera, etc, should be handled...
def check_instance(obj: Any, type_: Any) -> bool:
    """This function checks if an object is an instance of a given type. It supports generic types as well.

    :param obj: The object to check.
    :param type_: The type to check against. This can be a generic type like List[int] or Dict[str, Any].
    :return: True if the object is an instance of the type, False otherwise.
    """
    if type_ == Any:
        return True
    # Get the origin of the type (i.e., the base class for generic types)
    origin = getattr(type_, "__origin__", None)

    # If the type has an origin, it's a generic type
    if origin is not None:
        # If the type is a Union type
        if origin is Union:
            return any(check_instance(obj, t) for t in type_.__args__)
        elif origin is Literal:
            return obj in type_.__args__
        # Check if the object is an instance of the origin of the type
        elif not isinstance(obj, origin):
            return False

        # If the type has arguments (i.e., it's a parameterized generic type like List[int])
        if hasattr(type_, "__args__"):
            # Get the element type(s) of the generic type
            element_type = type_.__args__

            # If the object is a dictionary
            if isinstance(obj, dict):
                all_items_meet_condition = True

                # Iterate over each key-value pair in the dictionary
                for key, value in obj.items():
                    # Check if the key is an instance of the first element type and the value is an instance of the second element type
                    key_is_correct_type = check_instance(key, element_type[0])
                    value_is_correct_type = check_instance(value, element_type[1])

                    # If either the key or the value is not the correct type, set the flag to False and break the loop
                    if not key_is_correct_type or not value_is_correct_type:
                        all_items_meet_condition = False
                        break

                # Return the result
                return all_items_meet_condition

            # If the object is a list, set, or tuple
            elif isinstance(obj, (list, set, tuple)):
                element_type = element_type[0]
                for i in obj:
                    if not check_instance(i, element_type):
                        return False
                return True

    # If the type is not a generic type, just use isinstance
    return isinstance(obj, type_)



import logging

from hamilton.plugins.h_polars import PolarsDataFrameResult

logger = logging.getLogger(__name__)
logger.warning("This module is deprecated. Please use hamilton.plugins.h_polars instead.")

__all__ = ["PolarsDataFrameResult"]



import inspect

import pytest

from hamilton import ad_hoc_utils, function_modifiers


def test_copy_func():
    """Tests that we copy the function as intended"""

    @function_modifiers.tag(test_function="true")
    def foo(bar: int) -> int:
        """dummy function"""
        return bar + 1

    cloned_func = ad_hoc_utils._copy_func(foo)

    assert cloned_func.__dict__ == foo.__dict__
    assert cloned_func.__annotations__ == foo.__annotations__

    assert cloned_func is not foo
    assert cloned_func(1) == foo(1)


def test_generate_unique_temp_module_name():
    """Tests that we replace - with _"""
    name = ad_hoc_utils._generate_unique_temp_module_name()
    assert "-" not in name


def test_create_temporary_module():
    """Tests that we create a module with the passed in functions."""

    def bar(baz: int) -> int:
        """dummy function"""
        return baz + 1

    def foo(bar: int) -> int:
        """dummy function"""
        return bar + 1

    def _baz(bar: int) -> int:
        """dummy function, not to be included"""
        return bar + 1

    temp_module = ad_hoc_utils.create_temporary_module(bar, foo)
    expected_members = {
        "__spec__",
        "__loader__",
        "__name__",
        "__doc__",
        "bar",
        "foo",
        "__package__",
    }
    assert set(dict(inspect.getmembers(temp_module)).keys()) == expected_members
    assert "_" in temp_module.__name__
    temp_module_2 = ad_hoc_utils.create_temporary_module(bar, foo, module_name="test_module")
    assert set(dict(inspect.getmembers(temp_module_2)).keys()) == expected_members
    assert temp_module_2.__name__ == "test_module"


def test_create_temporary_module_breaks_helper():
    """Tests that we create a module with the passed in functions."""

    def bar(baz: int) -> int:
        """dummy function"""
        return baz + 1

    def foo(bar: int) -> int:
        """dummy function"""
        return bar + 1

    def _baz(bar: int) -> int:
        """dummy function, not to be included"""
        return bar + 1

    with pytest.raises(ValueError):
        ad_hoc_utils.create_temporary_module(bar, foo, _baz)


def test_inspect_module_from_source():
    source = '''
def bar(baz: int) -> int:
    """dummy function"""
    return baz + 1

def foo(bar: int) -> int:
    """dummy function"""
    return bar + 1

def _baz(bar: int) -> int:
    """dummy function, not to be included"""
    return bar + 1
'''
    module = ad_hoc_utils.module_from_source(source)

    try:
        inspect.getsource(module.bar)
    except OSError as e:
        assert False, f"module improperly added to linecache. {e}"



import logging
import pprint
from typing import Any, Collection, Dict, List, Optional, Set, Tuple

from hamilton import node
from hamilton.lifecycle.base import LifecycleAdapterSet

logger = logging.getLogger(__name__)

"""A set of utility functions for managing/traversing DAGs. Note these all operate on nodes.
We will likely want to genericize them so we're dealing with anything, not just node.Nodes.
"""


def topologically_sort_nodes(nodes: List[node.Node]) -> List[node.Node]:
    """Topologically sorts a list of nodes based on their dependencies.
    Note that we bypass utilizing the preset dependencies/depended_on_by attributes of the node,
    as we may want to use this before these nodes get put in a function graph.

    Thus we compute our own dependency map...
    Note that this assumes that the nodes are continuous -- if there is a hidden dependency that
    connects them, this has no way of knowing about it.

    TODO -- use python graphlib when we no longer have to support 3.8.

    https://docs.python.org/3/library/graphlib.html

    :param nodes: Nodes to sort
    :return: Nodes in sorted order
    """
    node_name_map = {node_.name: node_ for node_ in nodes}
    depended_on_by_map = {}
    dependency_map = {}
    for node_ in nodes:
        dependency_map[node_.name] = []
        for dep in node_.input_types:
            # if the dependency is not here, we don't want to count it
            # that means it depends on something outside the set of nodes we're sorting
            if dep not in node_name_map:
                continue
            dependency_map[node_.name].append(dep)
            if dep not in depended_on_by_map:
                depended_on_by_map[dep] = []
            depended_on_by_map[dep].append(node_)

    in_degrees = {node_.name: len(dependency_map.get(node_.name, [])) for node_ in nodes}
    # TODO -- determine what happens if nodes have dependencies that aren't present
    sources = [node_ for node_ in nodes if in_degrees[node_.name] == 0]
    queue = []
    for source in sources:
        queue.append(source)
    sorted_nodes = []
    while len(queue) > 0:
        node_ = queue.pop(0)
        sorted_nodes.append(node_)
        for next_node in depended_on_by_map.get(node_.name, []):
            if next_node.name in in_degrees:
                in_degrees[next_node.name] -= 1
                if in_degrees[next_node.name] == 0:
                    queue.append(next_node)
    return sorted_nodes


def get_node_levels(topologically_sorted_nodes: List[node.Node]) -> Dict[str, int]:
    """Gets the levels for a group of topologically sorted nodes.
    This only works if its topologically sorted, of course...


    :param topologically_sorted_nodes:
    :return: A dictionary of node name -> level
    """
    node_levels = {}
    node_set = {node_.name for node_ in topologically_sorted_nodes}
    for node_ in topologically_sorted_nodes:
        dependencies_in_set = {n.name for n in node_.dependencies}.intersection(node_set)
        if len(dependencies_in_set) == 0:
            node_levels[node_.name] = 0
        else:
            node_levels[node_.name] = max([node_levels[n] for n in dependencies_in_set]) + 1
    return node_levels


def combine_config_and_inputs(config: Dict[str, Any], inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Validates and combines config and inputs, ensuring that they're mutually disjoint.
    :param config: Config to construct, run the DAG with.
    :param inputs: Inputs to run the DAG on at runtime
    :return: The combined set of inputs to the DAG.
    :raises ValueError: if they are not disjoint
    """
    duplicated_inputs = [key for key in inputs if key in config]
    if len(duplicated_inputs) > 0:
        raise ValueError(
            f"The following inputs are present in both config and inputs. They must be "
            f"mutually disjoint. {duplicated_inputs} "
        )
    return {**config, **inputs}


def create_input_string(kwargs: dict) -> str:
    """This is a utility function to create a string representation of the inputs to a function.

    This is useful for debugging, as it can be printed out to see what the inputs were.

    :param kwargs: The inputs to the function that errored.
    :return: The string representation of the inputs, truncated appropriately.
    """
    pp = pprint.PrettyPrinter(width=80)
    inputs = {}
    for k, v in kwargs.items():
        item_repr = repr(v)
        if len(item_repr) > 50:
            item_repr = item_repr[:50] + "..."
        else:
            item_repr = v
        inputs[k] = item_repr
    input_string = pp.pformat(inputs)
    if len(input_string) > 1000:
        input_string = input_string[:1000] + "..."
    return input_string


def create_error_message(kwargs: dict, node_: node.Node, step: str) -> str:
    """Creates an error message for a node that errored."""
    # This code is coupled to how @config resolution works. Ideally it shouldn't be,
    # so when @config resolvers are changed to return Nodes, then fn.__name__ should
    # just work.
    original_func_name = "unknown"
    if node_.originating_functions:
        if hasattr(node_.originating_functions[0], "__original_name__"):
            original_func_name = node_.originating_functions[0].__original_name__
        else:
            original_func_name = node_.originating_functions[0].__name__
    module = (
        node_.originating_functions[0].__module__
        if node_.originating_functions and hasattr(node_.originating_functions[0], "__module__")
        else "unknown_module"
    )
    message = f">{step} {node_.name} [{module}.{original_func_name}()] encountered an error"
    padding = " " * (80 - min(len(message), 79) - 1)
    message += padding + "<"
    input_string = create_input_string(kwargs)
    message += "\n> Node inputs:\n" + input_string
    border = "*" * 80
    message = "\n" + border + "\n" + message + "\n" + border
    return message


def execute_subdag(
    nodes: Collection[node.Node],
    inputs: Dict[str, Any],
    adapter: LifecycleAdapterSet = None,
    computed: Dict[str, Any] = None,
    overrides: Dict[str, Any] = None,
    run_id: str = None,
    task_id: str = None,
) -> Dict[str, Any]:
    """Base function to execute a subdag. This conducts a depth first traversal of the graph.

    :param nodes: Nodes to compute
    :param inputs: Inputs, external
    :param adapter:  Adapter to use to compute
    :param computed:  Already computed nodes
    :param overrides: Overrides to use, will short-circuit computation
    :param run_id: Run ID to use
    :param task_id: Task ID to use -- this is optional for the purpose of the task-based execution...
    :return: The results
    """
    if overrides is None:
        overrides = {}
    if computed is None:
        computed = {}
    nodes_to_compute = {node_.name for node_ in nodes}

    if adapter is None:
        adapter = LifecycleAdapterSet()

    def dfs_traverse(
        node_: node.Node, dependency_type: node.DependencyType = node.DependencyType.REQUIRED
    ):
        if node_.name in computed:
            return
        if node_.name in overrides:
            computed[node_.name] = overrides[node_.name]
            return
        for n in node_.dependencies:
            if n.name not in computed:
                _, node_dependency_type = node_.input_types[n.name]
                dfs_traverse(n, node_dependency_type)

        logger.debug(f"Computing {node_.name}.")
        if node_.user_defined:
            if node_.name not in inputs:
                if dependency_type != node.DependencyType.OPTIONAL:
                    raise NotImplementedError(
                        f"{node_.name} was expected to be passed in but was not."
                    )
                return
            result = inputs[node_.name]
        else:
            kwargs = {}  # construct signature
            for dependency in node_.dependencies:
                if dependency.name in computed:
                    kwargs[dependency.name] = computed[dependency.name]
            error = None
            result = None
            success = True
            pre_node_execute_errored = False
            try:
                if adapter.does_hook("pre_node_execute", is_async=False):
                    try:
                        adapter.call_all_lifecycle_hooks_sync(
                            "pre_node_execute",
                            run_id=run_id,
                            node_=node_,
                            kwargs=kwargs,
                            task_id=task_id,
                        )
                    except Exception as e:
                        pre_node_execute_errored = True
                        raise e

                if adapter.does_method("do_node_execute", is_async=False):
                    result = adapter.call_lifecycle_method_sync(
                        "do_node_execute",
                        run_id=run_id,
                        node_=node_,
                        kwargs=kwargs,
                        task_id=task_id,
                    )
                else:
                    result = node_(**kwargs)
            except Exception as e:
                success = False
                error = e
                step = "[pre-node-execute]" if pre_node_execute_errored else ""
                message = create_error_message(kwargs, node_, step)
                logger.exception(message)
                raise
            finally:
                if not pre_node_execute_errored and adapter.does_hook(
                    "post_node_execute", is_async=False
                ):
                    try:
                        adapter.call_all_lifecycle_hooks_sync(
                            "post_node_execute",
                            run_id=run_id,
                            node_=node_,
                            kwargs=kwargs,
                            success=success,
                            error=error,
                            result=result,
                            task_id=task_id,
                        )
                    except Exception:
                        message = create_error_message(kwargs, node_, "[post-node-execute]")
                        logger.exception(message)
                        raise

        computed[node_.name] = result
        # > pruning the graph
        # This doesn't narrow it down to the entire space of the graph
        # E.G. if something is not needed by this current execution due to
        # the selection of nodes to run it might not prune everything.
        # to do this we'd need to first determine all nodes on the path, then prune
        # We may also want to use a reference counter for slightly cleaner/more efficient memory management

        for dep in node_.dependencies:
            if dep.name in computed and dep.name not in nodes_to_compute:
                for downstream_node in dep.depended_on_by:
                    # if it isn't computed, and it isn't required, we can't prune
                    if (
                        downstream_node.name not in computed
                        or downstream_node.name in nodes_to_compute
                    ):
                        break
                # If the result of this node is no longer needed, we can prune it/save the memory
                else:
                    del computed[dep.name]

    for final_var_node in nodes:
        dep_type = node.DependencyType.REQUIRED
        if final_var_node.user_defined:
            # from the top level, we don't know if this UserInput is required. So mark as optional.
            dep_type = node.DependencyType.OPTIONAL
        dfs_traverse(final_var_node, dep_type)
    return computed


def nodes_between(
    end_node: node.Node,
    search_condition: lambda node_: bool,
) -> Tuple[Optional[node.Node], List[node.Node]]:
    """Utility function to search backwards from an end node to a start node.
    This returns all nodes for which both of the following conditions are met:

    1. It contains a node that matches the start_condition as an ancestor
    2. It contains a node that matches the end node as a dependent

    Note that currently it is assumed that only one node will
    match search_condition.

    This just grabs the search node when it finds it -- the nonlocal is a bit hacky but more fun
    than passing a ton of data back and forth (who the parent is, etc...).

    :param end_node: Node to trace back from
    :param search_condition: Condition to stop the search for ancestors
    :return: A tuple of [start_node, between], where start_node is None
    if there is no path (and between will be empty).
    """

    out = set()
    visited = set()
    search_node = None

    def dfs_traverse(node_: node.Node):
        if search_condition(node_):
            # if we hit the end, we want to include all others in it
            nonlocal search_node
            search_node = node_
            return True
        if node_ in visited:
            # if we've already seen it, we want to include it
            return node_ in out
        # now we mark that we've seen it
        visited.add(node_)

        any_deps_included = False
        for n in node_.dependencies:
            any_deps_included |= dfs_traverse(n)
        if any_deps_included:
            out.add(node_)
        return any_deps_included

    for dep in end_node.dependencies:
        dfs_traverse(dep)

    return search_node, list(out)


def node_is_required_by_anything(node_: node.Node, node_set: Set[node.Node]) -> bool:
    """Checks dependencies on this node and determines if at least one requires it.

    Nodes can be optionally depended upon, i.e. the function parameter has a default value. We want to check that
    of the nodes the depend on this one, at least one of them requires it, i.e. the parameter is not optional.

    :param node_: node in question
    :param node_set: checks that we traverse only nodes in the provided set.
    :return: True if it is required by any downstream node, false otherwise
    """
    required = False
    for downstream_node in node_.depended_on_by:
        if downstream_node not in node_set:
            continue
        _, dep_type = downstream_node.input_types[node_.name]
        if dep_type == node.DependencyType.REQUIRED:
            return True
    return required



def json_to_save_1() -> dict:
    return {
        "key1": "value1",
        "key2": "value2",
    }


def json_to_save_2() -> dict:
    return {
        "key3": "value3",
        "key4": "value4",
    }


def expects_loader(external: dict) -> dict:
    return external



import pandas as pd

from hamilton.function_modifiers import ResolveAt, extract_columns, parameterize, resolve, source


@extract_columns("a", "b", "c", "d", "e")
def df() -> pd.DataFrame:
    """Produces a dataframe with columns a, b, c, d, e consisting entirely of 1s"""
    return pd.DataFrame(
        {
            "a": [1],
            "b": [1],
            "c": [1],
            "d": [1],
            "e": [1],
        }
    )


@resolve(
    when=ResolveAt.CONFIG_AVAILABLE,
    decorate_with=lambda columns_to_sum_map: parameterize(
        **{
            key: {"col_1": source(value[0]), "col_2": source(value[1])}
            for key, value in columns_to_sum_map.items()
        }
    ),
)
def generic_summation(col_1: pd.Series, col_2: pd.Series) -> pd.Series:
    return col_1 + col_2


# TODO -- add grouping then we can test everything



import inspect
import pathlib

import pytest

from hamilton import ad_hoc_utils, driver, graph_types, node
from hamilton.plugins import h_diskcache


def _callable_to_node(callable) -> node.Node:
    return node.Node(
        name=callable.__name__,
        typ=inspect.signature(callable).return_annotation,
        callabl=callable,
    )


@pytest.fixture()
def hook(tmp_path: pathlib.Path):
    return h_diskcache.DiskCacheAdapter(cache_path=str(tmp_path.resolve()))


@pytest.fixture()
def node_a():
    """Default function implementation"""

    def A(external_input: int) -> int:
        return external_input % 7

    return _callable_to_node(A)


@pytest.fixture()
def node_a_body():
    """The function A() has modulo 5 instead of 7"""

    def A(external_input: int) -> int:
        return external_input % 5

    return _callable_to_node(A)


@pytest.fixture()
def node_a_docstring():
    """The function A() has a docstring"""

    def A(external_input: int) -> int:
        """This one has a docstring"""
        return external_input % 7

    return _callable_to_node(A)


def test_set_result(hook: h_diskcache.DiskCacheAdapter, node_a: node.Node):
    """Hook sets value and assert value in cache"""
    node_hash = graph_types.hash_source_code(node_a.callable, strip=True)
    node_kwargs = dict(external_input=7)
    cache_key = (node_hash, *node_kwargs.values())
    result = 2

    hook.cache_vars = [node_a.name]
    # used_nodes_hash would be set by run_to_execute() hook
    hook.used_nodes_hash[node_a.name] = node_hash
    hook.run_after_node_execution(
        node_name=node_a.name,
        node_kwargs=node_kwargs,
        result=result,
    )

    # run_to_execute_node() hook would get cache
    assert hook.cache.get(key=cache_key) == result


def test_get_result(hook: h_diskcache.DiskCacheAdapter, node_a: node.Node):
    """Hooks get value and assert cache hit"""
    node_hash = graph_types.hash_source_code(node_a.callable, strip=True)
    node_kwargs = dict(external_input=7)
    result = 2
    cache_key = (node_hash, *node_kwargs.values())

    hook.cache_vars = [node_a.name]
    # run_after_node_execution() would set cache
    hook.cache.set(key=cache_key, value=result)
    hook.cache.stats(enable=True, reset=True)
    hook.run_to_execute_node(
        node_name=node_a.name,
        node_kwargs=node_kwargs,
        node_callable=node_a.callable,
    )

    # 1 hit, 0 miss
    assert hook.cache.stats() == (1, 0)


def test_append_nodes_history(
    hook: h_diskcache.DiskCacheAdapter,
    node_a: node.Node,
    node_a_body: node.Node,
):
    """Assert the CacheHook.nodes_history is growing;
    doesn't check for commit to cache
    """
    node_name = "A"
    node_kwargs = dict(external_input=7)
    hook.cache_vars = [node_a.name]

    # node version 1
    hook.used_nodes_hash[node_name] = graph_types.hash_source_code(node_a.callable, strip=True)
    hook.run_to_execute_node(
        node_name=node_name,
        node_kwargs=node_kwargs,
        node_callable=node_a.callable,
    )

    # check history
    assert len(hook.nodes_history.get(node_name, [])) == 1

    # node version 2
    hook.used_nodes_hash[node_name] = graph_types.hash_source_code(node_a_body.callable, strip=True)
    hook.run_to_execute_node(
        node_name=node_name,
        node_kwargs=node_kwargs,
        node_callable=node_a_body.callable,
    )

    assert len(hook.nodes_history.get(node_name, [])) == 2


def test_commit_nodes_history(hook: h_diskcache.DiskCacheAdapter):
    """Commit node history to cache"""
    hook.nodes_history = dict(A=["hash_1", "hash_2"])

    hook.run_after_graph_execution()

    assert hook.cache.get(h_diskcache.DiskCacheAdapter.nodes_history_key) == hook.nodes_history


def test_evict_all_except(
    hook: h_diskcache.DiskCacheAdapter,
    node_a: node.Node,
    node_a_body: node.Node,
):
    """Check utility function to evict all except passed nodes"""
    node_a_hash = graph_types.hash_source_code(node_a.callable, strip=True)
    node_a_body_hash = graph_types.hash_source_code(node_a_body.callable, strip=True)
    hook.cache[h_diskcache.DiskCacheAdapter.nodes_history_key] = dict(
        A=[node_a_hash, node_a_body_hash]
    )

    eviction_counter = h_diskcache.evict_all_except(nodes_to_keep=dict(A=node_a), cache=hook.cache)

    assert eviction_counter == 1
    assert hook.cache[h_diskcache.DiskCacheAdapter.nodes_history_key] == dict(A=[node_a_hash])


def test_evict_from_driver(
    hook: h_diskcache.DiskCacheAdapter,
    node_a: node.Node,
    node_a_body: node.Node,
):
    """Check utility function to evict all except driver"""
    node_a_hash = graph_types.hash_source_code(node_a.callable, strip=True)
    node_a_body_hash = graph_types.hash_source_code(node_a_body.callable, strip=True)
    hook.cache[h_diskcache.DiskCacheAdapter.nodes_history_key] = dict(
        A=[node_a_hash, node_a_body_hash]
    )
    module = ad_hoc_utils.create_temporary_module(node_a.callable)
    dr = driver.Builder().with_modules(module).with_adapters(hook).build()

    output = h_diskcache.evict_all_except_driver(dr)

    assert output["eviction_counter"] == 1
    assert hook.cache[h_diskcache.DiskCacheAdapter.nodes_history_key] == dict(A=[node_a_hash])
